#
# Copyright (C) 2020 Sylvain Marsat.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with with program; see the file COPYING. If not, write to the
#  Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#  MA  02111-1307  USA
#

"""
    Standalone python implementation of TaylorF2 phasing.
"""

import numpy as np
import lisabeta
import lisabeta.tools.pytools as pytools
import lisabeta.pyconstants as pyconstants

PN_order_max = 7

# c.f. notebook TF2.nb, end of section "Checks for SEOBNRv4T"
# NOTE: constants used (pi, gamma) are numpy constants here
def tf2_phasing_coeffs(q, chi1, chi2, kappa1=1., kappa2=2., PNorderNS=7, PNorderSO=7, PNorderSS=6):
    coeffs = {}

    eta = pytools.etaofq(q)
    delta = pytools.deltaofq(q)
    chi_s = 1./2 * (chi1 + chi2)
    chi_a = 1./2 * (chi1 - chi2)
    kappa_s = 1./2 * (kappa1 + kappa2)
    kappa_a = 1./2 * (kappa1 - kappa2)

    # Maximal orders
    coeffs['PNorderNS'] = PNorderNS
    coeffs['PNorderSO'] = PNorderSO
    coeffs['PNorderSS'] = PNorderSS

    # Newtonian coefficient, prefactor
    coeffs['prefactor'] = 3./128/eta

    # Non-spinning PN coefficients
    coeffs['NS'] = {}
    coeffs['NS'][0] = 1.
    coeffs['NS'][1] = 0.
    coeffs['NS'][2] = eta*(55/9.) + (3715/756.)
    coeffs['NS'][3] = -16*np.pi
    coeffs['NS'][4] = eta*eta*(3085/72.) + eta*(27145/504.) + (15293365/508032.)
    coeffs['NS'][5] = eta*np.pi*(-65/9.) + np.pi*(38645/756.)
    coeffs['NS'][6] = eta*(-15737765635/3048192.) + eta*eta*eta*(-127825/1296.) + np.log(2)*(-13696/21.) + np.euler_gamma*(-6848/21.) + np.pi*np.pi*(-640/3.) + eta*np.pi*np.pi*(2255/12.) + eta*eta*(76055/1728.) + (11583231236531/4694215680.)
    coeffs['NS'][7] = eta*eta*np.pi*(-74045/756.) + eta*np.pi*(378515/1512.) + np.pi*(77096675/254016.)
    # Non-spinning PN coefficients with factor log(v)
    coeffs['NSlogv'] = {}
    coeffs['NSlogv'][0] = 0.
    coeffs['NSlogv'][1] = 0.
    coeffs['NSlogv'][2] = 0.
    coeffs['NSlogv'][3] = 0.
    coeffs['NSlogv'][4] = 0.
    coeffs['NSlogv'][5] = eta*np.pi*(-65/3.) + np.pi*(38645/252.)
    coeffs['NSlogv'][6] = (-6848/21.)
    coeffs['NSlogv'][7] = 0.

    # Spin-orbit PN coefficients
    coeffs['SO'] = {}
    coeffs['SO'][0] = 0.
    coeffs['SO'][1] = 0.
    coeffs['SO'][2] = 0.
    coeffs['SO'][3] = chi_a*delta*(113/3.) + chi_s*(eta*(-76/3.) + (113/3.))
    coeffs['SO'][4] = 0.
    coeffs['SO'][5] = chi_a*(delta*(-732985/2268.) + delta*eta*(-140/9.)) + chi_s*((-732985/2268.) + eta*eta*(340/9.) + eta*(24260/81.))
    coeffs['SO'][6] = chi_a*delta*np.pi*(2270/3.) + chi_s*(-520*eta*np.pi + np.pi*(2270/3.))
    coeffs['SO'][7] = chi_a*(delta*(-25150083775/3048192.) + delta*eta*eta*(-1985/48.) + delta*eta*(26804935/6048.)) + chi_s*((-25150083775/3048192.) + eta*eta*(-1042165/3024.) + eta*eta*eta*(5345/36.) + eta*(10566655595/762048.))
    # Spin-orbit PN coefficients with factor log(v)
    coeffs['SOlogv'] = {}
    coeffs['SOlogv'][0] = 0.
    coeffs['SOlogv'][1] = 0.
    coeffs['SOlogv'][2] = 0.
    coeffs['SOlogv'][3] = 0.
    coeffs['SOlogv'][4] = 0.
    coeffs['SOlogv'][5] = chi_a*(delta*(-732985/756.) + delta*eta*(-140/3.)) + chi_s*((-732985/756.) + eta*eta*(340/3.) + eta*(24260/27.))
    coeffs['SOlogv'][6] = 0.
    coeffs['SOlogv'][7] = 0.

    # Spin-square PN coefficients
    coeffs['SS'] = {}
    coeffs['SS'][0] = 0.
    coeffs['SS'][1] = 0.
    coeffs['SS'][2] = 0.
    coeffs['SS'][3] = 0.
    coeffs['SS'][4] = chi_a*chi_s*(-100*kappa_a + 200*eta*kappa_a - 100*delta*kappa_s + delta*(-5/4.)) + chi_a*chi_a*(100*eta - 50*delta*kappa_a - 50*kappa_s + 100*eta*kappa_s + (-5/8.)) + chi_s*chi_s*(-50*delta*kappa_a - 50*kappa_s + 100*eta*kappa_s + eta*(-195/2.) + (-5/8.))
    coeffs['SS'][5] = 0.
    coeffs['SS'][6] = chi_a*chi_s*(-480*eta*eta*kappa_a + delta*(-1344475/1008.) + eta*kappa_a*(-88510/21.) + delta*eta*kappa_s*(-1495/3.) + delta*eta*(745/18.) + kappa_a*(26015/14.) + delta*kappa_s*(26015/14.)) + chi_a*chi_a*(-240*eta*eta - 240*eta*eta*kappa_s + (-1344475/2016.) + eta*kappa_s*(-44255/21.) + delta*eta*kappa_a*(-1495/6.) + delta*kappa_a*(26015/28.) + kappa_s*(26015/28.) + eta*(267815/252.)) + chi_s*chi_s*(-240*eta*eta*kappa_s + (-1344475/2016.) + eta*kappa_s*(-44255/21.) + delta*eta*kappa_a*(-1495/6.) + eta*eta*(3415/9.) + delta*kappa_a*(26015/28.) + kappa_s*(26015/28.) + eta*(829705/504.))
    # Spin-square PN coefficients with factor log(v)
    coeffs['SSlogv'] = {}
    coeffs['SSlogv'][0] = 0.
    coeffs['SSlogv'][1] = 0.
    coeffs['SSlogv'][2] = 0.
    coeffs['SSlogv'][3] = 0.
    coeffs['SSlogv'][4] = 0.
    coeffs['SSlogv'][5] = 0.
    coeffs['SSlogv'][6] = 0.

    # Precompute sum of NS, SO, SS coeffs
    coeffs['v'] = np.zeros(PN_order_max+1, dtype=float)
    coeffs['vlogv'] = np.zeros(PN_order_max+1, dtype=float)
    for i in range(PNorderNS+1):
        coeffs['v'][i] = coeffs['NS'][i]
        coeffs['vlogv'][i] = coeffs['NSlogv'][i]
    for i in range(PNorderSO+1):
        coeffs['v'][i] += coeffs['SO'][i]
        coeffs['vlogv'][i] += coeffs['SOlogv'][i]
    for i in range(PNorderSS+1):
        coeffs['v'][i] += coeffs['SS'][i]
        coeffs['vlogv'][i] += coeffs['SSlogv'][i]

    return coeffs

# The solar mass is from pyconstants
def tf2_phasing_eval(coeffs, M, f):

    M_s = M * pyconstants.MTSUN_SI

    v = np.power(np.pi * M_s * f, 1./3)
    logv = np.log(v)

    v_minus5 = np.power(v, -5.)
    v_n = {}
    v_n[1] = v
    for i in range(2, PN_order_max+1):
        v_n[i] = v * v_n[i-1]

    # Leading term -- for GR this is 1., coeffs['v'][0]=1. and coeffs['vlogv'][0]=0.
    phase_scaled = (coeffs['v'][0] + coeffs['vlogv'][0]*logv)
    for i in range(1, PN_order_max+1):
        phase_scaled += (coeffs['v'][i] + coeffs['vlogv'][i]*logv) * v_n[i]
    phase = coeffs['prefactor'] * v_minus5 * phase_scaled

    return phase

def tf2_phasing_eval_deriv(coeffs, M, f):

    M_s = M * pyconstants.MTSUN_SI

    v = np.power(np.pi * M_s * f, 1./3)
    logv = np.log(v)

    v_minus8 = np.power(v, -8.)
    v_n = {}
    v_n[0] = 1.
    for i in range(1, PN_order_max+1):
        v_n[i] = v * v_n[i-1]

    # Leading term -- for GR this is -5., coeffs['v'][0]=1. and coeffs['vlogv'][0]=0.
    dfphase_scaled = (((-5)*coeffs['v'][0] + coeffs['vlogv'][0]) + (-5)*coeffs['vlogv'][0]*logv)
    for i in range(1,PN_order_max+1):
        dfphase_scaled += (((i-5)*coeffs['v'][i] + coeffs['vlogv'][i]) + (i-5)*coeffs['vlogv'][i]*logv) * v_n[i]

    dfphase = np.pi*M_s/3. * coeffs['prefactor'] * v_minus8 * dfphase_scaled

    return dfphase

def tf2_phasing_eval_phase_and_deriv(coeffs, M, f):

    M_s = M * pyconstants.MTSUN_SI

    v = np.power(np.pi * M_s * f, 1./3)
    logv = np.log(v)

    v_minus8 = np.power(v, -8.)
    v_n = {}
    v_n[0] = 1.
    for i in range(1, PN_order_max+1):
        v_n[i] = v * v_n[i-1]

    # Leading term -- for GR this is 1., coeffs['v'][0]=1. and coeffs['vlogv'][0]=0.
    phase_scaled = (coeffs['v'][0] + coeffs['vlogv'][0]*logv)
    for i in range(1, PN_order_max+1):
        phase_scaled += (coeffs['v'][i] + coeffs['vlogv'][i]*logv) * v_n[i]
    phase = coeffs['prefactor'] * v_minus5 * phase_scaled

    # Leading term -- for GR this is -5., coeffs['v'][0]=1. and coeffs['vlogv'][0]=0.
    dfphase_scaled = (((-5)*coeffs['v'][0] + coeffs['vlogv'][0]) + (-5)*coeffs['vlogv'][0]*logv)
    for i in range(1,PN_order_max+1):
        dfphase_scaled += (((i-5)*coeffs['v'][i] + coeffs['vlogv'][i]) + (i-5)*coeffs['vlogv'][i]*logv) * v_n[i]
    dfphase = np.pi*M_s/3. * coeffs['prefactor'] * v_minus8 * dfphase_scaled

    return phase, dfphase

def tf2_phasing(freqs, M, q, chi1, chi2, kappa1=1., kappa2=2., f0=None, t0=0., phi0=0., PNorderNS=7, PNorderSO=7, PNorderSS=6):

    if f0 is None or f0<=0.:
        f0 = freqs[0]

    coeffs = tf2_phasing_coeffs(q, chi1, chi2, kappa1=kappa1, kappa2=kappa2, PNorderNS=PNorderNS, PNorderSO=PNorderSO, PNorderSS=PNorderSS)

    phi_f0 = tf2_phasing_eval(coeffs, M, f0)
    dfphi_f0 = tf2_phasing_eval_deriv(coeffs, M, f0)

    deltaphi_align = 2*np.pi*(t0 - dfphi_f0/(2*np.pi))*(freqs - f0) + phi0 - phi_f0

    phi = tf2_phasing_eval(coeffs, M, freqs) + deltaphi_align

    return phi

def tf2_phasing_tf(freqs, M, q, chi1, chi2, kappa1=1., kappa2=2., f0=None, t0=0., phi0=0., PNorderNS=7, PNorderSO=7, PNorderSS=6):

    if f0 is None or f0<=0.:
        f0 = freqs[0]

    coeffs = tf2_phasing_coeffs(q, chi1, chi2, kappa1=kappa1, kappa2=kappa2, PNorderNS=PNorderNS, PNorderSO=PNorderSO, PNorderSS=PNorderSS)

    phi_f0 = tf2_phasing_eval(coeffs, M, f0)
    dfphi_f0 = tf2_phasing_eval_deriv(coeffs, M, f0)

    deltaphi_align = 2*np.pi*(t0 - dfphi_f0/(2*np.pi))*(freqs - f0) + phi0 - phi_f0
    dfdeltaphi_align = 2*np.pi*(t0 - dfphi_f0/(2*np.pi))

    phi, dfphi = tf2_phasing_eval_and_deriv(coeffs, M, freqs)

    phi = phi + deltaphi_align
    tf = (dfphi + dfdeltaphi_align) / (2*np.pi)

    return phi, tf

# NOTE: for now, only Newtonian amplitude implemented
# DL in Mpc
def tf2_amp(freqs, M, q, chi1, chi2, DL, kappa1=1., kappa2=2., PNorderNS=7, PNorderSO=7, PNorderSS=6):

    if PNorderNS>0 or PNorderSO>0 or PNorderSS>0:
        raise ValueError("TF2 amplitude only implemented at Newtonian order for now.")

    eta = pytools.etaofq(q)
    M_s = M * pyconstants.MTSUN_SI
    M_m = M * pyconstants.MRSUN_SI

    v = np.power(np.pi * M_s * freqs, 1./3)

    aN = M_m / (DL * 1e6*pyconstants.PC_SI) * M_s * np.pi * np.sqrt(2*eta/3.) * np.power(v, -7./2)

    return aN



#
# Copyright (C) 2019 Sylvain Marsat.
#
#


"""
    Cython definitions for spline interpolation tools.
"""

import numpy as np
cimport numpy as np

from lisabeta.struct.pystruct cimport *
from lisabeta.tools.pyspline cimport *

cdef extern from "overlap.h":

    double complex ComputeFresnelIntegral(
        CAmpPhaseFDData* integrand
    );

    double complex FDSinglemodeFresnelOverlap(
        CAmpPhaseFDData* freqseries1,
        CAmpPhaseFDSpline* splines2,
        double fLow,
        double fHigh
    );

    double complex FDSinglemodeFresnelOverlap3Chan(
        CAmpPhaseFDData* freqseries1chan1,
        CAmpPhaseFDData* freqseries1chan2,
        CAmpPhaseFDData* freqseries1chan3,
        CAmpPhaseFDSpline* splines2chan1,
        CAmpPhaseFDSpline* splines2chan2,
        CAmpPhaseFDSpline* splines2chan3,
        double fLow,
        double fHigh
    );



#
# Copyright (C) 2019 Sylvain Marsat.
#
#


"""
    Cython wrapping of C LISA FD response.
"""

import numpy as np
cimport numpy as np
from enum import Enum

import lisabeta.pyconstants as pyconstants
import lisabeta.tools.pytools as pytools
import lisabeta.tools.pyspline as pyspline


################################################################################
# Structures and defaults
################################################################################

class TDIEnum(Enum):
    TDIAET = 0
    TDIXYZ = 1
    TDI2AET = 2
    TDI2XYZ = 3

class ResponseApproxEnum(Enum):
    full         = 0
    lowfL        = 1
    lowf         = 2
    ignoreRdelay = 3
    lowfL_highfsens = 4

LISAconst2010 = {
  'OrbitOmega' : _EarthOrbitOmega_SI,
  'OrbitPhi0' : 0.,
  'Orbitt0' : 0.,
  'OrbitR' : _AU_SI,
  'OrbitL' : 5e9
}

LISAconstProposal = {
  'OrbitOmega' : _EarthOrbitOmega_SI,
  'OrbitPhi0' : 0.,
  'Orbitt0' : 0.,
  'OrbitR' : _AU_SI,
  'OrbitL' : 2.5e9
}

LISAconstDict = {
  '2010': LISAconst2010,
  'Proposal': LISAconstProposal
}

################################################################################
# LISA TDI response
################################################################################

cdef class LISAFDresponseTDI3Chan:
    """ Transfer functions for the LISA response.
    """

    cdef real_vector* CphaseRdelay  # pointer to C vector, orbital delay phase
    cdef complex_vector* Ctransfer1 # pointer to C vector, channel 1 transfer
    cdef complex_vector* Ctransfer2 # pointer to C vector, channel 2 transfer
    cdef complex_vector* Ctransfer3 # pointer to C vector, channel 3 transfer

    cdef freq                       # input numpy array for frequencies
    cdef tf                         # input numpy array for tf spline

    cdef real_vector* Cfreq         # pointer to C frequency vector
    cdef real_vector* Ctf           # pointer to C vector for tf

    cdef double t0                  # Reference orbital time (yr)
    cdef int l                      # Mode number l
    cdef int m                      # Mode number m
    cdef double inc                 # Inclination in source-frame
    cdef double phi                 # Azimuthal phase in source-frame
    cdef double lambd               # Ecliptic longitude, SSB-frame
    cdef double beta                # Ecliptic latitude, SSB-frame
    cdef double psi                 # Polarization angle, SSB-frame

    cdef LISAconst                  # dict for LISA orbital constants
    cdef responseapprox             # tag for approximation in LISA response
    cdef TDI                        # choice of TDI observables
    cdef frozenLISA                 # tag to treat LISA as motionless
    cdef TDIrescaled                # Flag to rescale TDI variables

    cdef phaseRdelay                # numpy array, orbital delay phase
    cdef transfer1                  # numpy array, channel 1 transfer
    cdef transfer2                  # numpy array, channel 2 transfer
    cdef transfer3                  # numpy array, channel 3 transfer


    def __init__(self,
                 np.ndarray[np.float_t, ndim=1] freq,
                 np.ndarray[np.float_t, ndim=1] tf,
                 t0, l, m, inc, phi, lambd, beta, psi,
                 TDI='TDIAET', LISAconst=LISAconstProposal,
                 responseapprox='full', frozenLISA=False, TDIrescaled=False):
        """Constructor
        Args:
          freq                       # input numpy array for frequencies
          tf                         # input numpy array for tf

          t0                  # Reference orbital time (yr)
          l                   # Mode number l
          m                   # Mode number m
          inc                 # Inclination in source-frame
          phi                 # Azimuthal phase in source-frame
          lambd               # Ecliptic longitude, SSB-frame
          beta                # Ecliptic latitude, SSB-frame
          psi                 # Polarization angle, SSB-frame
        Keyword args:
          TDI                 # Choice of TDI variables
          LISAconst           # Choice of LISA constellation
          responseapprox      # Choice of low-f approximation in the response
          frozenLISA          # Flag to freeze LISA motion
          TDIrescaled         # Flag to apply rescaling to TDI variables
        """

        self.freq = freq
        self.tf = tf

        self.t0 = t0
        self.l = l
        self.m = m
        self.inc = inc
        self.phi = phi
        self.lambd = lambd
        self.beta = beta
        self.psi = psi

        self.TDI = TDI
        self.responseapprox = responseapprox
        self.frozenLISA = frozenLISA
        self.TDIrescaled = TDIrescaled
        if isinstance(LISAconst, basestring):
            self.LISAconst = LISAconstDict[LISAconst]
        else:
            self.LISAconst = LISAconst

        self.Cfreq = NULL
        self.Ctf = NULL
        self.CphaseRdelay = NULL
        self.Ctransfer1 = NULL
        self.Ctransfer2 = NULL
        self.Ctransfer3 = NULL

        # If the input is an empty array, init to empty arrays and stop there
        # NOTE: __init__ is required to return nothing
        if len(freq)==0 or len(tf)==0:
            self.phaseRdelay = np.empty(0, dtype=float)
            self.transfer1 = np.empty(0, dtype=float)
            self.transfer2 = np.empty(0, dtype=float)
            self.transfer3 = np.empty(0, dtype=float)
            return

        # Build a real_vector representation of the input numpy arrays
        if not freq.flags['C_CONTIGUOUS']:
            raise ValueError('Input numpy array freq is not C_CONTIGUOUS')
        if not tf.flags['C_CONTIGUOUS']:
            raise ValueError('Input numpy array tf is not C_CONTIGUOUS')
        cdef double* freq_data = <double *> &freq[0]
        self.Cfreq = real_vector_view(freq_data, freq.shape[0])

        cdef double* tf_data = <double *> &tf[0]
        self.Ctf = real_vector_view(tf_data, tf.shape[0])

        cdef CTDI = TDIEnum[self.TDI].value
        cdef Cresponseapprox = ResponseApproxEnum[self.responseapprox].value
        cdef int CfrozenLISA = <int> self.frozenLISA
        cdef int Crescaled = <int> self.TDIrescaled

        cdef LISAconstellation CLISAconst
        CLISAconst.OrbitOmega = self.LISAconst['OrbitOmega']
        CLISAconst.OrbitPhi0 = self.LISAconst['OrbitPhi0']
        CLISAconst.Orbitt0 = self.LISAconst['Orbitt0']
        CLISAconst.OrbitR = self.LISAconst['OrbitR']
        CLISAconst.OrbitL = self.LISAconst['OrbitL']

        ret = EvalLISAFDresponseTDI3Chan(
            &self.CphaseRdelay,
            &self.Ctransfer1,
            &self.Ctransfer2,
            &self.Ctransfer3,
            self.Cfreq,
            self.Ctf,
            self.t0,
            self.l,
            self.m,
            self.inc,
            self.phi,
            self.lambd,
            self.beta,
            self.psi,
            CTDI,
            &CLISAconst,
            Cresponseapprox,
            CfrozenLISA,
            Crescaled
        );
        if ret == _FAILURE:
            raise ValueError("Call to EvalLISAFDresponseTDI3Chan() failed.")

        # Cast C double array to numpy via a MemoryView
        cdef double[::1] view_phaseRdelay = \
          <(double)[:self.CphaseRdelay.size]> self.CphaseRdelay.data
        self.phaseRdelay = np.asarray(view_phaseRdelay)
        cdef double complex[::1] view_transfer1 = \
          <(double complex)[:self.Ctransfer1.size]> self.Ctransfer1.data
        self.transfer1 = np.asarray(view_transfer1)
        cdef double complex[::1] view_transfer2 = \
          <(double complex)[:self.Ctransfer2.size]> self.Ctransfer2.data
        self.transfer2 = np.asarray(view_transfer2)
        cdef double complex[::1] view_transfer3 = \
          <(double complex)[:self.Ctransfer3.size]> self.Ctransfer3.data
        self.transfer3 = np.asarray(view_transfer3)
        #self.Mfreq = real_vector_to_np_array(self.CMfreq)

    def __dealloc__(self):
        """Destructor
        """
        if self.CphaseRdelay != NULL:
            real_vector_free(self.CphaseRdelay)
        if self.Ctransfer1 != NULL:
            complex_vector_free(self.Ctransfer1)
        if self.Ctransfer2 != NULL:
            complex_vector_free(self.Ctransfer2)
        if self.Ctransfer3 != NULL:
            complex_vector_free(self.Ctransfer3)
        if self.Cfreq != NULL:
            real_vector_view_free(self.Cfreq)
        if self.Ctf != NULL:
            real_vector_view_free(self.Ctf)

    def get_response(self):
        return np.copy(self.phaseRdelay), np.copy(self.transfer1), \
                 np.copy(self.transfer2), np.copy(self.transfer3),

################################################################################
# Signal modulations from orbit around AGN
################################################################################

def chirp_bands(t, phi, fdot_exclude_mono=2e-12):
    spline_phi = pyspline.CubicSpline(t, phi)
    phiddot = spline_phi.get_spline_dd()[:,1]
    # We put a minus sign here so as to give the + sign to chirping domains
    # Indeed, \ddot{phi_{22}} is negative since phi_{22} = -2 phi_{orb}
    # The 2pi is just to convert to frequency f instead of omega
    x, y = t, -phiddot / (2*np.pi)
    interval = [-fdot_exclude_mono, fdot_exclude_mono]
    t_bands = pytools.func_domain_decomp_linear(x, y, interval)
    return t_bands

def chirping_signal_decomp(signal_td, t_bands, window_fraction=0.25, npt_min=20):
    signals = []
    # TODO: building splines probably done outside this function already ?
    spline_ampClass = pyspline.CubicSpline(signal_td[:,0], signal_td[:,1])
    spline_phiClass = pyspline.CubicSpline(signal_td[:,0], signal_td[:,2])
    spline_amp = spline_ampClass.get_spline()
    spline_phi = spline_phiClass.get_spline()
    for i, band in enumerate(t_bands):
        if band[2]==0:
            deltat = (band[1] - band[0]) * window_fraction
            deltatf = deltat
            deltati = deltat
            ti = band[0] + deltat
            tf = band[1] - deltat
        else:
            if i>0:
                deltati = window_fraction * (t_bands[i-1][1] - t_bands[i-1][0])
                ti = band[0] - deltati
            else:
                deltati = 0.
                ti = band[0]
            if i<len(t_bands)-1:
                deltatf = window_fraction * (t_bands[i+1][1] - t_bands[i+1][0])
                tf = band[1] + deltatf
            else:
                deltatf = 0.
                tf = band[1]
        s_r = pytools.restrict_data_bycol(signal_td, [ti, tf], col=0)
        if len(s_r)<npt_min:
            times = np.linspace(ti, tf, npt_min)
            s_r = np.zeros((npt_min, 3), dtype=float)
            s_r[:,0] = times
            s_r[:,1] = pyspline.spline_eval_vector(spline_amp, times)
            s_r[:,2] = pyspline.spline_eval_vector(spline_phi, times)
        w = pytools.window_planck_vec(s_r[:,0], ti, tf, deltati, deltatf)
        s_w = np.array([s_r[:,0], s_r[:,1]*w, s_r[:,2]]).T
        signals += [s_w]
    return signals

# NOTE: convention tr = te + d(te)
# NOTE: by convention, d(t0) = 0 (t0 in s)
# NOTE: R_AGN in units of M_AGN (e.g. R=700M), M_AGN in solar masses
def doppler_modulated_signal_td(signal_td, t0=0., **AGNparams):
    t, amp, phase = signal_td.T
    shapiro_delay = AGNparams.get('shapiro_delay', False)
    M_AGN = AGNparams['M_AGN']
    R_AGN = AGNparams['R_AGN']
    theta_AGN = AGNparams['theta_AGN']
    phi_AGN = AGNparams['phi_AGN']
    M_AGN_s = M_AGN * pyconstants.MTSUN_SI
    Omega_AGN = 1./(M_AGN_s) * (R_AGN**(-3/2.))
    # Leading order geometric propagation delay
    d = -R_AGN * M_AGN_s * np.sin(theta_AGN) * np.cos(Omega_AGN*t - phi_AGN)
    d0 = -R_AGN * M_AGN_s * np.sin(theta_AGN) * np.cos(Omega_AGN*t0 - phi_AGN)
    # Shapiro time delay if included, with a constant scaled out by convention
    d_shapiro = np.zeros_like(d0)
    d0_shapiro = np.zeros_like(d0)
    if shapiro_delay:
        d_shapiro = -2*M_AGN_s * np.log(1 + np.sin(theta_AGN) * np.cos(Omega_AGN*t - phi_AGN))
        d0_shapiro = -2*M_AGN_s * np.log(1 + np.sin(theta_AGN) * np.cos(Omega_AGN*t0 - phi_AGN))
    td = t + d - d0 + d_shapiro - d0_shapiro
    return np.array([td, amp, phase]).T



#
# Copyright (C) 2019 Sylvain Marsat.
#
#


"""
    Cython wrapping of numerical constants.
"""

import numpy as np
cimport numpy as np


from lisabeta.pyconstants cimport *

SUCCESS = _SUCCESS
FAILURE = _FAILURE

PI = _PI
PI_2 = _PI_2
PI_3 = _PI_3
PI_4 = _PI_4
SQRTPI = _SQRTPI
SQRTTWOPI = _SQRTTWOPI
INVSQRTPI = _INVSQRTPI
INVSQRTTWOPI = _INVSQRTTWOPI
GAMMA = _GAMMA
SQRT2 = _SQRT2
SQRT3 = _SQRT3
SQRT6 = _SQRT6
INVSQRT2 = _INVSQRT2
INVSQRT3 = _INVSQRT3
INVSQRT6 = _INVSQRT6

C_SI = _C_SI
G_SI = _G_SI
MSUN_SI = _MSUN_SI
MTSUN_SI = _MTSUN_SI
MRSUN_SI = _MRSUN_SI
PC_SI = _PC_SI
AU_SI = _AU_SI
YRSID_SI = _YRSID_SI

EarthOrbitOmega_SI = _EarthOrbitOmega_SI

EPOCH_J2000_0_TAI_UTC = _EPOCH_J2000_0_TAI_UTC
EPOCH_J2000_0_GPS = _EPOCH_J2000_0_GPS
EPOCH_LISA_0_GPS = _EPOCH_LISA_0_GPS



#
#  Copyright (C) 2019 Sylvain Marsat.
#


"""
    Cython wrapping of LISA noise

    The underlying C functions work on one value of f at a time.
    Interface for a numpy array of freq values is provided as a python loop.
"""

# from lisabeta.pyconstants cimport *
# from lisabeta.struct.pystruct cimport *
cimport numpy as np
import numpy as np
from enum import Enum
import lisabeta.pyconstants as pyconstants
import lisabeta.tools.pyspline as pyspline
from pyLISAnoise cimport *
import lisabeta.lisa.pyresponse as pyresponse
from pyresponse cimport *
from pyresponse import LISAconst2010, LISAconstProposal, TDIEnum
#cimport lisabeta.struct.pystruct as pystruct

class LISANoiseEnum(Enum):
    LISASciRDv1noise  = 0
    LISAProposalnoise = 1
    LISA2010noise     = 2
    LISA2017noise     = 3

LISAnoiseDict = {
  'SciRDv1': LISANoiseEnum.LISASciRDv1noise.value,
  'Proposal': LISANoiseEnum.LISAProposalnoise.value,
  '2010': LISANoiseEnum.LISA2010noise.value,
  '2017': LISANoiseEnum.LISA2017noise.value
}

LISAnoiseSciRDv1 = {
  'InstrumentalNoise' : LISANoiseEnum.LISASciRDv1noise.value,
  'WDbackground' : True,
  'WDduration' : 3.,
  'lowf_add_pm_noise_f0': 0.,
  'lowf_add_pm_noise_alpha': 2.
}

LISAnoiseProposal = {
  'InstrumentalNoise' : LISANoiseEnum.LISAProposalnoise.value,
  'WDbackground' : True,
  'WDduration' : 3.,
  'lowf_add_pm_noise_f0': 0.,
  'lowf_add_pm_noise_alpha': 2.
}

LISAnoise2017 = {
  'InstrumentalNoise' : LISANoiseEnum.LISA2017noise.value,
  'WDbackground' : False,
  'WDduration' : 3.,
  'lowf_add_pm_noise_f0': 0.,
  'lowf_add_pm_noise_alpha': 2.
}

cdef class LISANoisePSDFunction(py_object_function):

    def __init__(self, LISAnoise, int nchan,
                 basestring TDI='TDIAET', TDIrescaled=True,
                 LISAconst=LISAconstProposal):
        """Constructor
        Args:
          LISAnoise           # Choice of noise model
          nchan               # Choice of TDI channel id (1,2,3)
        Keyword args:
          TDI                 # Choice of TDI variables
          LISAconst           # Choice of LISA constellation
          TDIrescaled         # Flag to apply rescaling to TDI variables
        """

        # Set up orbital constants from dictionary
        self.CLISAconst.OrbitOmega = LISAconst['OrbitOmega']
        self.CLISAconst.OrbitPhi0 = LISAconst['OrbitPhi0']
        self.CLISAconst.Orbitt0 = LISAconst['Orbitt0']
        self.CLISAconst.OrbitR = LISAconst['OrbitR']
        self.CLISAconst.OrbitL = LISAconst['OrbitL']

        # Set up noise model from dictionary
        instrumentalnoise = LISAnoise['InstrumentalNoise']
        if isinstance(instrumentalnoise, basestring):
            self.CLISAnoise.noise = LISAnoiseDict[instrumentalnoise]
        else:
            self.CLISAnoise.noise = instrumentalnoise
        self.CLISAnoise.WDbackground = int(LISAnoise['WDbackground'])
        self.CLISAnoise.WDduration = LISAnoise['WDduration']
        self.CLISAnoise.lowf_add_pm_noise_f0 = LISAnoise['lowf_add_pm_noise_f0']
        self.CLISAnoise.lowf_add_pm_noise_alpha = LISAnoise['lowf_add_pm_noise_alpha']

        # Set up noise function
        self.fn = NULL
        self.CLISAdata = LISA_noise_func_data(&self.CLISAnoise, &self.CLISAconst)
        CTDI = TDIEnum[TDI].value
        if (CTDI==TDIEnum['TDIAET'].value):
            if TDIrescaled:
                if nchan==1: self.fn = SnA
                elif nchan==2: self.fn = SnE
                elif nchan==3: self.fn = SnT
                else: raise ValueError("No such channel: "+str(nchan))
            else:
                if nchan==1: self.fn = SnANoRescaling
                elif nchan==2: self.fn = SnENoRescaling
                elif nchan==3: self.fn = SnTNoRescaling
                else: raise ValueError("No such channel: "+str(nchan))
        elif (CTDI==TDIEnum['TDIXYZ'].value):
            if TDIrescaled:
                self.fn = SnXYZ
            else:
                self.fn = SnXYZNoRescaling
        elif (CTDI==TDIEnum['TDI2AET'].value):
            if TDIrescaled:
                if nchan==1: self.fn = SnA
                elif nchan==2: self.fn = SnE
                elif nchan==3: self.fn = SnT
                else: raise ValueError("No such channel: "+str(nchan))
            else:
                if nchan==1: self.fn = SnA2NoRescaling
                elif nchan==2: self.fn = SnE2NoRescaling
                elif nchan==3: self.fn = SnT2NoRescaling
                else: raise ValueError("No such channel: "+str(nchan))
        elif (CTDI==TDIEnum['TDI2XYZ'].value):
            if TDIrescaled:
                self.fn = SnXYZ
            else:
                self.fn = SnXYZ2NoRescaling
        else: raise ValueError("No support for TDI = '"+str(nchan)+"'")

        self.obfn.object = <void*> &self.CLISAdata
        self.obfn.function = self.fn

    # cpdef double call(self, double f):
    #     return self.fn(&self.CLISAdata, f)
    #
    # cpdef np.ndarray[np.float_t, ndim=1] apply(self,
    #                                        np.ndarray[np.float_t, ndim=1] freq):
    #     cdef np.ndarray[np.float_t, ndim=1] psd = np.zeros_like(freq)
    #     for i in range(len(freq)):
    #         psd[i] = self.fn(&self.CLISAdata, freq[i])
    #     return psd


cdef class LISANoisePSDFunction3Chan:

    def __init__(self, LISAnoise,
                 basestring TDI='TDIAET', TDIrescaled=True,
                 LISAconst=LISAconstProposal):
        """Constructor
        Args:
          LISAnoise           # Choice of noise model
        Keyword args:
          TDI                 # Choice of TDI variables
          TDIrescaled         # Flag to apply rescaling to TDI variables
          LISAconst           # Choice of LISA constellation
        """

        # Set up orbital constants from dictionary
        if isinstance(LISAconst, basestring):
            LISAconst = pyresponse.LISAconstDict[LISAconst]
        self.CLISAconst.OrbitOmega = LISAconst['OrbitOmega']
        self.CLISAconst.OrbitPhi0 = LISAconst['OrbitPhi0']
        self.CLISAconst.Orbitt0 = LISAconst['Orbitt0']
        self.CLISAconst.OrbitR = LISAconst['OrbitR']
        self.CLISAconst.OrbitL = LISAconst['OrbitL']

        # Set up noise model from dictionary
        instrumentalnoise = LISAnoise['InstrumentalNoise']
        if isinstance(instrumentalnoise, basestring):
            self.CLISAnoise.noise = LISAnoiseDict[instrumentalnoise]
        else:
            self.CLISAnoise.noise = instrumentalnoise
        self.CLISAnoise.WDbackground = int(LISAnoise['WDbackground'])
        self.CLISAnoise.WDduration = LISAnoise['WDduration']
        self.CLISAnoise.lowf_add_pm_noise_f0 = LISAnoise['lowf_add_pm_noise_f0']
        self.CLISAnoise.lowf_add_pm_noise_alpha = LISAnoise['lowf_add_pm_noise_alpha']

        # Set up noise function
        self.fn = NULL
        self.CLISAdata = LISA_noise_func_data(&self.CLISAnoise, &self.CLISAconst)
        CTDI = TDIEnum[TDI].value
        if (CTDI==TDIEnum['TDIAET'].value):
            if TDIrescaled:
                self.fn = Sn3ChanAET
            else:
                self.fn = Sn3ChanAETNoRescaling
        elif (CTDI==TDIEnum['TDIXYZ'].value):
            if TDIrescaled:
                self.fn = Sn3ChanXYZ
            else:
                self.fn = Sn3ChanXYZNoRescaling
        elif (CTDI==TDIEnum['TDI2AET'].value):
            if TDIrescaled:
                self.fn = Sn3ChanAET
            else:
                self.fn = Sn3ChanAET2NoRescaling
        elif (CTDI==TDIEnum['TDI2XYZ'].value):
            if TDIrescaled:
                self.fn = Sn3ChanXYZ
            else:
                self.fn = Sn3ChanXYZ2NoRescaling
        else: raise ValueError("No support for TDI = " + TDI +".")
#
#         # self.obfn.object = <void*> &self.CLISAdata
#         # self.obfn.function = self.fn
#

    cpdef np.ndarray[np.float_t, ndim=2] apply(self, np.ndarray[np.float_t, ndim=1] freqs):
        cdef np.ndarray[np.float_t, ndim=2] Sndata = np.zeros((len(freqs), 4))
        cdef double Snval1
        cdef double Snval2
        cdef double Snval3
        cdef double f
        for i in range(len(freqs)):
            f = freqs[i]
            self.fn(&Snval1, &Snval2, &Snval3, &self.CLISAdata, f)
            Sndata[i,0] = f
            Sndata[i,1] = Snval1
            Sndata[i,2] = Snval2
            Sndata[i,3] = Snval3
        return Sndata

# Written as a pure numpy function for now
# Only implemented for TDI AET
def rescale_tdi_noise(freq, Sn1, Sn2, Sn3, TDI='TDIAET', LISAconst=LISAconstProposal):
    if isinstance(LISAconst, basestring):
        LISAconst = pyresponse.LISAconstDict[LISAconst]
    L = LISAconst['OrbitL']
    if TDI=='TDIAET':
        pifL = np.pi * freq * L / pyconstants.C_SI
        rescalingAE = 2 * np.sin(2 * pifL)**2
        rescalingT = 8 * np.sin(2 * pifL)**2 * np.sin(pifL)**2
        return Sn1 / rescalingAE, Sn2 / rescalingAE, Sn3 / rescalingT
    elif TDI=='TDI2AET':
        pifL = np.pi * freq * L / pyconstants.C_SI
        factor_tdi2 = 4*np.sin(4 * pifL)**2
        rescalingAE = factor_tdi2 * 2 * np.sin(2 * pifL)**2
        rescalingT = factor_tdi2 * 8 * np.sin(2 * pifL)**2 * np.sin(pifL)**2
        return Sn1 / rescalingAE, Sn2 / rescalingAE, Sn3 / rescalingT
    else:
        raise ValueError('For now, only TDIAET and TDI2AET implemented.')

# Simple function to initialize noise evaluator
# Allows the normal analytical interface as well as numerical data as input
# In the latter case, we build interpolating spline here (to avoid repeating this step)
# TODO: giving a rescaled numerical input is not supported yet
def initialize_noise(LISAnoise, TDI='TDIAET',
                     TDIrescaled=True,
                     LISAconst=LISAconstProposal):
    if isinstance(LISAnoise, np.ndarray):
        freq_Sn = np.copy(LISAnoise[:,0])
        Sn1_data = np.copy(LISAnoise[:,1])
        Sn2_data = np.copy(LISAnoise[:,2])
        Sn3_data = np.copy(LISAnoise[:,3])
        Sn1_splineClass = pyspline.CubicSpline(freq_Sn, Sn1_data)
        Sn2_splineClass = pyspline.CubicSpline(freq_Sn, Sn2_data)
        Sn3_splineClass = pyspline.CubicSpline(freq_Sn, Sn3_data)
        Sn1_spline = Sn1_splineClass.get_spline()
        Sn2_spline = Sn2_splineClass.get_spline()
        Sn3_spline = Sn3_splineClass.get_spline()
        noise_evaluator = [Sn1_spline, Sn2_spline, Sn3_spline]
    else:
        SnAPSDFunc3ChanClass = LISANoisePSDFunction3Chan(
                                             LISAnoise, TDI=TDI,
                                             TDIrescaled=TDIrescaled,
                                             LISAconst=LISAconst)
        noise_evaluator = SnAPSDFunc3ChanClass
    return noise_evaluator

# Simple function to evaluate the noise
# Allows the normal analytical interface as well as numerical data as input
def evaluate_noise(LISAnoise, noise_evaluator, freq,
                   TDI='TDIAET', TDIrescaled=True, LISAconst=LISAconstProposal):
    if isinstance(LISAnoise, np.ndarray):
        Sn1_spline = noise_evaluator[0]
        Sn2_spline = noise_evaluator[1]
        Sn3_spline = noise_evaluator[2]
        Sn1_vals = pyspline.spline_eval_vector(Sn1_spline, freq)
        Sn2_vals = pyspline.spline_eval_vector(Sn2_spline, freq)
        Sn3_vals = pyspline.spline_eval_vector(Sn3_spline, freq)
        if TDIrescaled:
            Sn1_vals, Sn2_vals, Sn3_vals = rescale_tdi_noise(freq, Sn1_vals, Sn2_vals, Sn3_vals, TDI=TDI, LISAconst=LISAconst)
    else:
        Sndata = noise_evaluator.apply(freq)
        Sn1_vals = Sndata[:,1]
        Sn2_vals = Sndata[:,2]
        Sn3_vals = Sndata[:,3]

    return Sn1_vals, Sn2_vals, Sn3_vals



#
# Copyright (C) 2020 Sylvain Marsat.
#
#


"""
    Python functions for ptemcee simulated inference of SBHBs in AGN orbit with LISA.
"""


import os
import sys
import copy
import json
import h5py
import argparse
import numpy as np

import lisabeta
import lisabeta.pyconstants as pyconstants
import lisabeta.tools.pytools as pytools
import lisabeta.lisa.lisatools as lisatools
import lisabeta.lisa.snrtools as snrtools
import lisabeta.lisa.lisa as lisa
import lisabeta.lisa.lisa_fisher as lisa_fisher
import lisabeta.inference.inference as inference

import ptemcee
import ptemcee.mpi_pool as mpi_pool

try:
    from mpi4py import MPI
except ModuleNotFoundError:
    MPI = None


################################################################################
# Parse arguments, default params

# Parse arguments
parser = argparse.ArgumentParser(description='Run SBHB AGN parameter inference \
                                              using lisabeta and ptemcee.')
parser.add_argument('input_file', type=str, help='Input json file \
                                                  with all parameters.')
args = parser.parse_args()

# Full list of physical params - infer_params are a subset
# TODO: generalize to other choices of mass and spin parameters
# TODO: at the moment, infer_params can only be a subset of this, not flexible
list_params = [
    "Mchirp",
    "q",
    "chi1",
    "chi2",
    "fstart",
    "dist",
    "inc",
    "phi",
    "lambda",
    "beta",
    "psi",
    "Omega_AGN",
    "Rcostheta_AGN",
    "Rsintheta_AGN",
    "phi_AGN"]

# Default waveform params
waveform_params_default = {
    "minf": 1e-5,
    "maxf": 0.5,
    "fend": None,
    "tstart": 0.0,
    "tmin": None,
    "tmax": None,
    "phiref": 0.0,
    "fref_for_phiref": 0.0,
    "toffset": 0.0,
    "TDI": "TDIAET",
    "acc": 1e-4,
    "order_fresnel_stencil": 0,
    "approximant": "IMRPhenomD",
    "LISAconst": "Proposal",
    "responseapprox": "full",
    "frozenLISA": False,
    "TDIrescaled": True,
    "LISAnoise": {
    "InstrumentalNoise": "SciRDv1",
    "WDbackground": True,
    "WDduration" : 3.0
    },
    "shapiro_delay": True,
    "waveform_params_update_for_template": {}
}

# Default pymultinest params
# TODO: print_info, n_iter_info very rudimentary at the moment
run_params_musthave = ['out_dir', 'out_name']
run_params_default = {
    "sampler": "ptemcee",
    "sample_Lframe": False,
    "simple_likelihood": False,
    "multimodal": False,
    "multimodal_pattern": "1mode",
    "p_jump": 0.,
    "skip_fisher": True,
    "init_method": "prior",
    "init_file": None,
    "init_scale_cov": 1.,
    "zerolike": False,
    "n_temps": 5,
    "temp_max": None,
    "adaptive_temp": False,
    "n_walkers": 100,
    "n_iter": 1000,
    "burn_in": 100,
    "autocor_method": "acor",
    "thin_samples": True,
    "upsample": 1,
    "seed": None,
    "print_info": False,
    "n_iter_info": 10,
    "output": True,
    "output_raw": True
}

# Default parameter range for angular parameters
params_range_default = {
    "inc": [0., np.pi],
    "phi": [-np.pi, np.pi],
    "lambda": [-np.pi, np.pi],
    "beta": [-np.pi/2, np.pi/2],
    "psi": [0., np.pi],
    "phi_AGN": [-np.pi, np.pi]
}


################################################################################
# Main

if __name__ == '__main__':

    # Using MPI or not
    # We allow for different cases: MPI used or not, master or worker
    use_mpi = False
    is_master = True
    mapper = map
    if MPI is not None:
        MPI_size = MPI.COMM_WORLD.Get_size()
        MPI_rank = MPI.COMM_WORLD.Get_rank()
        use_mpi = (MPI_size > 1)
        if use_mpi:
            print("MPI rank/size: %d / %d" % (MPI_rank, MPI_size), flush=True)
            pool = ptemcee.mpi_pool.MPIPool(debug=False)
            is_master = pool.is_master()
            mapper = pool.map
        else:
            print("No MPI", flush=True)
            is_master = True
            mapper = map

    # Load json file with all parameters
    with open(args.input_file, 'r') as input_file:
        input_params = json.load(input_file)
    # Source params and prior params must be given
    source_params = input_params['source_params']
    source_params = pytools.complete_mass_params(source_params)
    prior_params = input_params['prior_params']
    # Source params can be specified in the Lframe
    source_params_are_Lframe = source_params.pop('Lframe', False)
    # Waveform params and run params have predefined default values
    # Need to cast from json to python, list of modes list[list] -> list[tuple]
    waveform_params = waveform_params_default.copy()
    input_params['waveform_params'] = inference.waveform_params_json2py(
                                                input_params['waveform_params'])
    waveform_params.update(input_params['waveform_params'])
    run_params = run_params_default.copy()
    run_params.update(input_params['run_params'])
    zerolike = run_params.pop('zerolike', False)
    sample_Lframe = run_params.pop('sample_Lframe', False)
    simple_likelihood = run_params.pop('simple_likelihood', False)

    # Check run params are meant to be for the ptemcee sampler
    if not (run_params['sampler'] == 'ptemcee'):
        raise ValueError('run_params sampler flag is %s instead of ptemcee.'\
                          % (run_params['sampler']))
    # run params must contain output prefix for files -- no default here
    for musthave in run_params_musthave:
        if not (musthave in run_params):
            raise ValueError('run_params must contain %s.' % (musthave))
    # Check that the output directory exists
    if (run_params['output'] or run_params['output_raw']):
        if not os.path.isdir(run_params['out_dir']):
            raise ValueError('Output dir %s does not exist.' % run_params['out_dir'])

    # Lframe not supported for SBHBs
    if source_params_are_Lframe or sample_Lframe:
        raise ValueError('L-frame not relevant/supported for SBHBs.')
    source_params_SSBframe = source_params.copy()

    # Use default param ranges to complete input ranges -- check completeness
    prior_type = prior_params['prior_type']
    params_range = prior_params['params_range']
    infer_params = prior_params['infer_params']
    for i, param in enumerate(infer_params):
        if params_range[i] == []:
            params_range[i] = params_range_default[param]
    if [] in params_range:
        raise ValueError('Some parameter ranges are neither given \
                      nor covered by the defaults.')

    # Parameters inferred and parameters fixed
    infer_params = prior_params['infer_params']
    fixed_params = [p for p in list_params if p not in infer_params]
    n_dim = len(infer_params)
    n_fixed = len(fixed_params)
    # Vector x of infer_params for the injection
    # Also store values of fixed params in dict for later use
    # if sample_Lframe:
    #     fixed_params_dict = dict([(p, source_params_Lframe[p]) for p in fixed_params])
    #     x_inj = np.array([source_params_Lframe[p] for p in infer_params])
    # else:
    fixed_params_dict = dict([(p, source_params_SSBframe[p]) for p in fixed_params])
    x_inj = np.array([source_params_SSBframe[p] for p in infer_params])

    # Wrapped parameters
    wrap_params = prior_params.get('wrap_params', None)
    if wrap_params is None:
        wrap_params = inference.func_default_wrap(infer_params)
    else:
        if np.array(wrap_params).dtype == 'bool':
            raise ValueError('wrap_params in outdated format, expecting None (for default) or a list of [None or [a,b]] with [a,b] the wrap moduli, typically [-np.pi, np.pi].')

    # Multimodal jumps in L-frame not relevant and not implemented for SBHBs
    if run_params['multimodal']:
        raise ValueError('Multimodal jumps in L-frame not relevant and not implemented for SBHBs.')
    # Simple likelihood
    if simple_likelihood:
        raise ValueError('Simple likelihood not relevant and not implemented for SBHBs.')

    # Prior function (assumes independent multiplicative prior in all params)
    def prior(x):
        p = 1.
        for i in range(n_dim):
            p *= inference.compute_prior(prior_type[i], params_range[i], x[i])
        return p
    # Set up prior bounds for x vector of parameters, dimension (n_dim, 2)
    prior_bounds = np.array(params_range)
    # Check the injection is within the prior bounds
    if not inference.prior_check(x_inj, prior_bounds):
        raise ValueError('Source parameters not within the prior bounds!')

    # Set up likelihood class
    likelihoodClass = lisa.LikelihoodLISASOBH_AGN(source_params_SSBframe,
                                                  **waveform_params)

    # Waveform params to be used for template waveforms
    # May differ from the waveform params for injection, given as an update
    waveform_params_template = copy.deepcopy(waveform_params)
    if 'waveform_params_update_for_template' in waveform_params:
        waveform_params_template.update(
                         waveform_params['waveform_params_update_for_template'])

    # Define ln-likelihood and ln-prior function, wrapped for use in ptemcee
    # TODO: add m1>m2 to prior range checks
    def lnlikelihood(x):
        if not inference.prior_check(x, prior_bounds):
            return -1e99
        template_params = {}
        for i in range(n_dim):
            template_params[infer_params[i]] = x[i]
        for p in fixed_params:
            template_params[p] = fixed_params_dict[p]
        template_params = pytools.complete_mass_params(template_params)
        template_params = pytools.complete_spin_params(template_params)
        template_params['Lframe'] = sample_Lframe
        try:
            lnL = likelihoodClass.lnL(template_params, **waveform_params_template)
        except:
            np.set_printoptions(precision=16)
            print("Likelihood encountered exception for following x:", flush=True, file=sys.stderr)
            print(repr(x), flush=True, file=sys.stderr)
            return -1e99
        if np.isnan(lnL):
            np.set_printoptions(precision=16)
            print("Likelihood returned nan for following x:", flush=True, file=sys.stderr)
            print(repr(x), flush=True, file=sys.stderr)
            return -1e99
        return lnL
    def lnprior(x):
        if not inference.prior_check(x, prior_bounds):
            return -1e99
        p = prior(x)
        if p==0.: return -1e99
        return np.log(p)

    # Read run_params
    # p_extra = run_params['p_jump']
    init_method = run_params['init_method']
    init_scale_cov = run_params['init_scale_cov']
    n_temps = run_params['n_temps']
    temp_max = run_params['temp_max']
    adaptive_temp = run_params['adaptive_temp']
    n_walkers = run_params['n_walkers']
    n_iter = run_params['n_iter']
    burn_in = run_params['burn_in']
    autocor_method = run_params['autocor_method']
    upsample = run_params['upsample']
    seed = run_params['seed']
    # Only master will print info
    print_info = run_params['print_info'] and is_master
    n_iter_info = run_params['n_iter_info']
    n_dim = len(infer_params)

    if print_info:
        print('Source parameters SSBframe:')
        print(source_params_SSBframe)
        # print('Source parameters Lframe:')
        # print(source_params_Lframe)

    # Fisher matrix not available for SBHBs yet
    if (not run_params['skip_fisher']):
        raise ValueError('Fisher not implemented for SBHBs yet.')

    # Initialize walkers, drawing from the prior or from Fisher matrix
    x_ini = np.zeros((n_temps, n_walkers, n_dim))
    if init_method=='prior':
        # Drawing from the prior -- assumes independent priors in all params
        for k in range(n_temps):
            for j in range(n_dim):
                x_ini[k,:,j] = inference.draw_prior(prior_type[j], prior_bounds[j], n_walkers)
    elif init_method=='fisher':
        raise ValueError('Fisher not implemented for SBHBs yet.')
        # if run_params['skip_fisher']:
        #     raise ValueError('Cannot have skip_fisher and init_method=fisher.')
        # # Drawing from the Fisher matrix - same for each temperature
        # if sample_Lframe:
        #     mean = np.array([source_params_Lframe[p] for p in infer_params])
        # else:
        #     mean = np.array([source_params_SSBframe[p] for p in infer_params])
        # # We allow for the possibility of scaling the covariance for initialization
        # cov = fishercov['cov'] * init_scale_cov
        # try:
        #     for k in range(n_temps):
        #         for i in range(n_walkers):
        #             x_ini[k,i,:] = inference.draw_multivariate_normal_constrained(
        #                                   mean, cov, params_range, infer_params)
        # except:
        #     print('Fisher initialization failed, initializing from prior.')
        #     # Drawing from the prior -- assumes independent priors in all params
        #     for k in range(n_temps):
        #         for j in range(n_dim):
        #             x_ini[k,:,j] = inference.draw_prior(prior_type[j], prior_bounds[j], n_walkers)
    elif init_method=='file':
        init_file = run_params['init_file']
        with h5py.File(init_file, 'r') as hf:
            for j in range(n_dim):
                p = infer_params[j]
                x_ini[:,:,j] = np.reshape(hf[p][:], (n_temps, n_walkers))
    else:
        raise ValueError('Option init_method = %s not recognized.' % init_method)

    # Initial points all start on the main mode (1,0)
    # if multimodal:
    #     x_ini[:,n_dim] = 1
    #     x_ini[:,n_dim+1] = 0

    # Enforce range of phase parameters in case Fisher uncertainties are large
    # TODO: we should have proper rejection based on the prior ranges
    if 'inc' in infer_params:
        iinc = infer_params.index('inc')
        x_ini[:,:,iinc] = pytools.modpi(x_ini[:,:,iinc])
    if 'phi' in infer_params:
        iphi = infer_params.index('phi')
        x_ini[:,:,iphi] = pytools.mod2pi(x_ini[:,:,iphi])
    if 'lambda' in infer_params:
        ilambda = infer_params.index('lambda')
        x_ini[:,:,ilambda] = pytools.mod2pi(x_ini[:,:,ilambda])
    if 'beta' in infer_params:
        ibeta = infer_params.index('beta')
        x_ini[:,:,ibeta] = pytools.modpi2(x_ini[:,:,ibeta])
    if 'psi' in infer_params:
        ipsi = infer_params.index('psi')
        x_ini[:,:,ipsi] = pytools.mod2pi(x_ini[:,:,ipsi])

    # Temperature ladder
    betas = ptemcee.sampler.make_ladder(n_dim, ntemps=n_temps, Tmax=temp_max)

    # If using mpi, workers are simply told to wait
    # They will be sollicited by master for likelihood computations via pool.map
    if use_mpi and (not is_master):
        pool.wait()
        # Is that exit necessary ? Was in the emcee/schwimmbad example
        # Maybe there to ensure exits if pool.close() is not called
        sys.exit(0)

    # Whether using mpi and being master or simply not using mpi
    # Sampler set-up, running, post-processing and output all done by master
    if is_master:
        # Set up and initialize sampler
        sampler = ptemcee.sampler.Sampler(n_walkers, n_dim, lnlikelihood, lnprior,
                                          betas=betas, mapper=mapper,
                                          adaptive=adaptive_temp,
                                          list_param_wrap=wrap_params)
        chain = sampler.chain(x_ini)

        # A la grace de dieu
        if print_info:
            print('Running ptemcee...')
        chain.run(n_iter)

        # TODO: output meta information, acceptance rate and likelihood speed

        # Post-processing: unfold if multimodal was used
        if print_info:
            print('Post-processing...')

        # Only keep 0-temperature chain for post-processing
        # TODO: allow to output other temperatures, and temp. adjustments, for debug
        # ptemcee format: (niter, ntemp, nwalker, ndim)
        # We translate to multiemcee format: (nwalker, niter, ndim) for temp=0
        chain_unfold = np.transpose(chain.x[:,0,:,:], axes=(1,0,2))
        # Same format without ndim for like and post values
        lnlikevals = np.transpose(chain.logl[:,0,:], axes=(1,0))
        lnpostvals = np.transpose(chain.logP[:,0,:], axes=(1,0))
        # Format: (niter, ntemp)
        betavals = chain.betas
        # ptemcee does not seem to store individual acceptance, only their count
        acceptance_ratios = chain.jump_acceptance_ratio[0,:]

        # Post-processing: burn-in, autocor. length and thinning, merge walkers
        # Also complete with params that were fixed to their values
        if burn_in>=n_iter:
            print('WARNING: burn_in is larger than n_iter, ignoring burn_in.')
            burn_in = 0
        if run_params['thin_samples']:
            autocor_len = {}
            for i,param in enumerate(infer_params):
                # Note: autocorr_new computes a mean across walkers
                autocor_len[param] = inference.get_autocor(chain_unfold[:,burn_in:,i], method=autocor_method)
                if print_info:
                    print('Autocorrelation length for %s: %g' % (param,autocor_len[param]))
            # The thinning length is the worse autocor length for all params
            thin_len = int(np.ceil(np.max([autocor_len[param] for param in infer_params])))
            thin_len = int(np.ceil(thin_len / float(upsample)))
        else:
            thin_len = 1
        if print_info:
            print('Autocorrelation thinning: %d' % (thin_len,))
        # Processed output: thinned and merged chain (completed with fixed params)
        n_out = n_walkers * ((n_iter-burn_in-1) // thin_len + 1)
        chain_processed = np.zeros((n_out, n_dim + n_fixed), dtype=float)
        # Here ravel with order='F' flattens by reading along columns first
        for i in range(n_dim):
            i_out = list_params.index(infer_params[i])
            chain_processed[:,i_out] = np.ravel(chain_unfold[:,burn_in::thin_len,i],
                                                order='F')
        for i in range(n_fixed):
            i_out = list_params.index(fixed_params[i])
            # if sample_Lframe:
            #     chain_processed[:,i_out] = source_params_Lframe[fixed_params[i]]
            # else:
            chain_processed[:,i_out] = source_params_SSBframe[fixed_params[i]]
        # Thin and merge likelihood, posterior, acceptance vals
        lnlikevals_processed = np.ravel(lnlikevals[:,burn_in::thin_len], order='F')
        lnpostvals_processed = np.ravel(lnpostvals[:,burn_in::thin_len], order='F')
        # acceptvals_processed = np.ravel(acceptvals[:,burn_in::thin_len], order='F')

        # Output
        basename = run_params['out_dir'] + run_params['out_name']
        # Main output: processed, thinned chains and Fisher matrix
        # TODO: hierarchy, create a level (group ?) samples
        if run_params['output']:
            with h5py.File(basename + '.h5', 'w') as f:
                # Source params
                # source_params_Lframe_gr = f.create_group('source_params_Lframe')
                # pytools.write_h5py_dict(source_params_Lframe_gr, source_params_Lframe)
                source_params_SSBframe_gr = f.create_group('source_params_SSBframe')
                pytools.write_h5py_dict(source_params_SSBframe_gr, source_params_SSBframe)
                for i,param in enumerate(list_params):
                    f.create_dataset(param, data=chain_processed[:,i])
                f.create_dataset('lnlike', data=lnlikevals_processed)
                f.create_dataset('lnpost', data=lnpostvals_processed)
                f.create_dataset('acceptance_ratios', data=acceptance_ratios)
                f.create_dataset('betas', data=betavals)
                # f.create_dataset('accept', data=acceptvals_processed)
                # if not run_params['skip_fisher']:
                #     fishercov_gr = f.create_group('fishercov')
                #     lisa_fisher.write_h5py_fishercov(fishercov_gr, fishercov)
        # Extra output: full chains, walker-separated, for testing and diagnostics
        if run_params['output_raw']:
            with h5py.File(basename + '_raw.h5', 'w') as f:
                for i,param in enumerate(infer_params):
                    f.create_dataset(param, data=chain_unfold[:,:,i])
                f.create_dataset('lnlike', data=lnlikevals)
                f.create_dataset('lnpost', data=lnpostvals)
                # f.create_dataset('accept', data=acceptvals)

        # If using mpi, close the pool
        if use_mpi:
            pool.close()



#
#  Copyright (C) 2019 Sylvain Marsat.
#


"""
    Cython declarations for LISA noise
"""

cimport numpy as np

# from lisabeta.pyconstants cimport *
# from lisabeta.struct.pystruct cimport *
from pyresponse cimport *

cdef extern from "LISAnoise.h":
    #double SnXYZ(const LISANoiseModel *LISAnoise,
    #             const LISAconstellation *LISAconst,
    #             double f);
    #double SnA(const LISANoiseModel *LISAnoise,
    #              const LISAconstellation *LISAconst,
    #              double f);
    #double SnE(const LISANoiseModel *LISAnoise,
    #              const LISAconstellation *LISAconst,
    #              double f);
    #double SnT(const LISANoiseModel *LISAnoise,
    #              const LISAconstellation *LISAconst,
    #              double f);
    #double SnXYZNoRescaling(const LISANoiseModel *LISAnoise,
    #                         const LISAconstellation *LISAconst,
    #                         double f);
    #double SnANoRescaling(const LISANoiseModel *LISAnoise,
    #                         const LISAconstellation *LISAconst,
    #                         double f);
    #double SnENoRescaling(const LISANoiseModel *LISAnoise,
    #                         const LISAconstellation *LISAconst,
    #                         double f);
    #double SnTNoRescaling(const LISANoiseModel *LISAnoise,
    #                         const LISAconstellation *LISAconst,
    #                         double f);

    double SnXYZ(const void* variant, double f);
    double SnA(const void* variant, double f);
    double SnE(const void* variant, double f);
    double SnT(const void* variant, double f);
    double SnXYZNoRescaling(const void* variant, double f);
    double SnANoRescaling(const void* variant, double f);
    double SnENoRescaling(const void* variant, double f);
    double SnTNoRescaling(const void* variant, double f);
    double SnXYZ2NoRescaling(const void* variant, double f);
    double SnA2NoRescaling(const void* variant, double f);
    double SnE2NoRescaling(const void* variant, double f);
    double SnT2NoRescaling(const void* variant, double f);

    void Sn3ChanXYZ(double* SnX,
                    double* SnY,
                    double* SnZ,
                    const void* object,
                    double f);

    void Sn3ChanXYZNoRescaling(double* SnX,
                               double* SnY,
                               double* SnZ,
                               const void* object,
                               double f);

    void Sn3ChanXYZ2NoRescaling(double* SnX,
                               double* SnY,
                               double* SnZ,
                               const void* object,
                               double f);

    void Sn3ChanAET(double* SnA,
                    double* SnE,
                    double* SnT,
                    const void* object,
                    double f);

    void Sn3ChanAETNoRescaling(double* SnA,
                               double* SnE,
                               double* SnT,
                               const void* object,
                               double f);

    void Sn3ChanAET2NoRescaling(double* SnA,
                               double* SnE,
                               double* SnT,
                               const void* object,
                               double f);

    enum LISAInstrumentalNoise:
        LISASciRDv1noise  = 0
        LISAProposalnoise = 1
        LISA2010noise     = 2
        LISA2017noise     = 3

    ctypedef struct LISANoiseModel:
        LISAInstrumentalNoise noise;
        int WDbackground;
        double WDduration;
        double lowf_add_pm_noise_f0;
        double lowf_add_pm_noise_alpha;

    ctypedef struct LISA_noise_func_data:
        const LISANoiseModel *LISAnoise;
        const LISAconstellation *LISAconst;



#ctypedef double (*CLISANoisePSDFunc)(const LISANoiseModel *LISAnoise,
#                                     const LISAconstellation *LISAconst,
#                                     double f)
ctypedef double (*CLISANoisePSDFunc)(const void* variant, double f)
ctypedef void (*CLISANoisePSDFunc3Chan)(double* Snval1, double* Snval2, double* Snval3, const void* variant, double f)
ctypedef int (*CLISANoisePSDFuncVec)(real_vector** Snvals, const void* variant, real_vector* freq)

cdef class LISANoisePSDFunction(py_object_function):
    cdef LISAconstellation CLISAconst
    cdef LISANoiseModel CLISAnoise
    cdef LISA_noise_func_data CLISAdata
    cdef CLISANoisePSDFunc fn
    #cdef ... fn_vec
    #cdef object_function obfn
    #cpdef double call(self, double f)
    #cpdef np.ndarray[np.float_t, ndim=1] apply(self, np.ndarray[np.float_t, ndim=1] freq)
    #cpdef np.ndarray[np.float_t, ndim=1] apply_vec(self, np.ndarray[np.float_t, ndim=1] freq)

cdef class LISANoisePSDFunction3Chan:
    cdef LISAconstellation CLISAconst
    cdef LISANoiseModel CLISAnoise
    cdef LISA_noise_func_data CLISAdata
    cdef CLISANoisePSDFunc3Chan fn

    cpdef np.ndarray[np.float_t, ndim=2] apply(self, np.ndarray[np.float_t, ndim=1] freqs)



#
# Copyright (C) 2019 Sylvain Marsat
#


"""
    Python plotting tools.
"""

from __future__ import absolute_import, division, print_function
import sys
if sys.version_info[0] == 2:
    from future_builtins import map, filter


import copy
import json
import h5py
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

import lisabeta.tools.pytools as pytools
import lisabeta.lisa.pyresponse as pyresponse
import lisabeta.lisa.lisatools as lisatools
import lisabeta.lisa.lisa_fisher as lisa_fisher
import lisabeta.utils.corner_covar as corner_covar


################################################################################
# Plotting functions and options
################################################################################

################################################################################
# Functions to plot arrays
# Input format :
# ax axes object (for subplot use)
# arg data sequence [data1, [colx1, coly1]], ...
# kwargs options : domain (figsize now at the level of subplot)
# allow for two data formats, distinguished by wether the 2nd arg in the list is a numpy array or not :
# i) [nparray d, cols [i,j]] : datax = d[:,i], datay = d[:,j]
# ii) [nparray datax, nparray datay] directly
# typical usage for continuous color map: [cm, cmb] = ['inferno', [0.,0.9]] lplot(...,  colormap=cm, colormapbounds=cmb)
# Color palette stolen from seaborn package (deep, with reshuffled ordering)
# SEABORN_PALETTES = dict(
#     deep=["#4C72B0", "#55A868", "#C44E52",
#           "#8172B2", "#CCB974", "#64B5CD"],
#     muted=["#4878CF", "#6ACC65", "#D65F5F",
#            "#B47CC7", "#C4AD66", "#77BEDB"],
#     pastel=["#92C6FF", "#97F0AA", "#FF9F9A",
#             "#D0BBFF", "#FFFEA3", "#B0E0E6"],
#     bright=["#003FFF", "#03ED3A", "#E8000B",
#             "#8A2BE2", "#FFC400", "#00D7FF"],
#     dark=["#001C7F", "#017517", "#8C0900",
#           "#7600A1", "#B8860B", "#006374"],
#     colorblind=["#0072B2", "#009E73", "#D55E00",
#                 "#CC79A7", "#F0E442", "#56B4E9"]
#)
# Added colors by hand:
# import matplotlib._color_data as mcd
# mcd.CSS4_COLORS['orange']
# '#FFA500'
# mcd.CSS4_COLORS['sienna']
# '#A0522D'
rc_params = {'backend': 'ps',
            'font.family': 'Times New Roman',
            'font.sans-serif': ['Bitstream Vera Sans'],
            'axes.unicode_minus':False,
            'text.usetex':True,
            'axes.grid':True,
            'grid.linestyle':':',
            'grid.linewidth':1.,
            'hist.bins':50,
            'axes.labelsize':16,
            'axes.titlesize':16,
            'xtick.labelsize':16,
            'ytick.labelsize':16,
            'legend.fontsize':16,
            'savefig.bbox':'tight',
            'savefig.transparent':True,
            'figure.dpi':300}
plt.rcParams.update(rc_params)

plotpalette = ["#4C72B0", "#C44E52", "#CCB974", "#55A868", "#8172B2", "#64B5CD", '#FFA500', '#A0522D']
def lplot(ax, *args, **kwargs):
    rangex = kwargs.pop('rangex', [])
    rangey = kwargs.pop('rangey', [])
    ds = kwargs.pop('downsample', 1)
    size = kwargs.pop('figsize', (8, 4))
    grid = kwargs.pop('grid', True)
    colormap = kwargs.pop('colormap', None)
    colormapbounds = kwargs.pop('colormapbounds', [0.,1.])
    colors = kwargs.pop('colors', None)
    linestyles = kwargs.pop('linestyles', None)
    linewidths = kwargs.pop('linewidths', None)
    markers = kwargs.pop('markers', None)
    labels = kwargs.pop('labels', None)
    log_xscale = kwargs.pop('log_xscale', False)
    log_yscale = kwargs.pop('log_yscale', False)
    n = len(args)
    if colors is None: # colors option supersedes colormap
        if colormap is not None:
            colorm = cm.get_cmap(colormap)
            colors = [colorm(x) for x in np.linspace(colormapbounds[0],
                                                     colormapbounds[1], n)]
        else:
            #defaultcolorlist = plt.rcParams['axes.prop_cycle'].by_key()['color']
            defaultcolorlist = plotpalette
            colors = pytools.fold_list(defaultcolorlist, n)
    if linestyles is None:
        linestyles = ['-']*n
    if linewidths is None:
        linewidths = [1]*n
    if markers is None:
        markers = [None]*n
    if labels is None:
        labels = [None]*n
    f = plt.figure(0, figsize=size)
    minxvals = np.zeros(n)
    maxxvals = np.zeros(n)
    minyvals = np.zeros(n)
    maxyvals = np.zeros(n)
    avyvals = np.zeros(n)
    for i, x in enumerate(args):
        if type(x[1]) is np.ndarray:
            data = pytools.restrict_data_soft(np.array([x[0][::ds], x[1][::ds]]).T, rangex)
            col1, col2 = [0, 1]
        else:
            data = pytools.restrict_data_soft(x[0][::ds], rangex)
            col1, col2 = x[1]
        if not (log_xscale and log_yscale):
            minxvals[i] = data[0, col1]
        else: # Restrict to the first non-zero value of y - convenient for log-x plots (also always exclude x=0)
            datax = data[:, col1]
            datay = data[:, col2]
            if datax[0]==0.:
                datax = datax[1:]
                datay = datay[1:]
            minxvals[i] = datax[(datay > 0)][0]
        maxxvals[i] = data[-1, col1]
        minyvals[i] = min(data[:, col2])
        maxyvals[i] = max(data[:, col2])
        avyvals[i] = np.average(data[:, col2])
        ax.plot(data[:,col1], data[:,col2], color=colors[i], linestyle=linestyles[i], linewidth=linewidths[i], marker=markers[i], label=labels[i], **kwargs)
    if rangex:
        ax.set_xlim(rangex[0], rangex[1])
    else:
        ax.set_xlim(min(minxvals), max(maxxvals))
    if rangey:
        ax.set_ylim(rangey[0], rangey[1])
    else:
        if log_yscale:
            minyvalplot = max(min(minyvals), 1e-8*np.average(avyvals))
            ax.set_ylim(1./2*minyvalplot, 2*max(maxyvals))
        else:
            if max(maxyvals)==min(minyvals): # Collapsed case: plot a constant, scale is arbitrary, just plot +-1.
                ax.set_ylim(min(minyvals) - 1., max(maxyvals) + 1.)
            else:
                margin = 0.1 * (max(maxyvals) - np.average(avyvals))
                ax.set_ylim(min(minyvals) - margin, max(maxyvals) + margin)
    if log_xscale:
        ax.set_xscale('log')
    if log_yscale:
        ax.set_yscale('log')
    if grid:
        ax.grid('on', linestyle=':')
def llogplot(ax, *arg, **kwargs):
    args = (ax,) + arg
    return lplot(*args, log_yscale=True, **kwargs)
def lloglinearplot(ax, *arg, **kwargs):
    args = (ax,) + arg
    return lplot(*args, log_xscale=True, **kwargs)
def lloglogplot(ax, *arg, **kwargs):
    args = (ax,) + arg
    return lplot(*args, log_xscale=True, log_yscale=True, **kwargs)

################################################################################
# Functions to parse and manipulate parameter files and posterior files
################################################################################

# Load posterior file
# Assumed format: param1 param2 ... paramn lnL, parameters read from .json
# Fixed parameters completed with injected value read from .json
def load_params_posterior_lisa_smbh(params_file, posterior_file, format='multiemcee', load_fisher=True):

    params_post = {}

    # Load input params
    with open(params_file, 'r') as input_file:
        input_params = json.load(input_file)
    source_params = input_params['source_params']
    waveform_params = input_params['waveform_params']
    run_params = input_params['run_params']
    source_params_are_Lframe = source_params.get('Lframe', False)
    prior_params = input_params['prior_params']
    t0 = waveform_params.get('t0', 0.)

    # If source params are given in the Lframe, convert to SSB
    if source_params_are_Lframe:
        source_params_SSBframe = lisatools.convert_Lframe_to_SSBframe(
                                    source_params,
                                    t0=t0,
                                    frozenLISA=waveform_params['frozenLISA'])
        source_params_Lframe = source_params.copy()
    else:
        source_params_SSBframe = source_params.copy()
        source_params_Lframe = lisatools.convert_SSBframe_to_Lframe(
                                    source_params,
                                    t0=t0,
                                    frozenLISA=waveform_params['frozenLISA'])
    source_params_SSBframe['t0'] = t0
    source_params_Lframe['t0'] = t0

    # Load posterior from file
    if format=='multinest':
        post_raw = np.genfromtxt(posterior_file)
        post = {}
        for i,p in enumerate(prior_params['infer_params']):
            post[p] = post_raw[:,i]
        post['lnL'] = post_raw[:,-1]
    elif format=='multiemcee':
        with h5py.File(posterior_file, 'r') as hf:
            post = {}
            for p in prior_params['infer_params']:
                post[p] = hf[p][:]
            post['lnL'] = hf['lnlike'][:]
            post['lnpost'] = hf['lnpost'][:]
    elif format=='multifisher':
        with h5py.File(posterior_file, 'r') as hf:
            post = {}
            for p in prior_params['infer_params']:
                post[p] = hf[p][:]
            post['lnL'] = hf['lnlike'][:]
            post['lnpost'] = hf['lnpost'][:]
            post['lnweight'] = hf['lnweight'][:]
    else:
        raise ValueError('Format not recognized')

    # For multiemcee, we also load fisher
    if not format=='multinest' and load_fisher:
        with h5py.File(posterior_file, 'r') as hf:
            params_post['fishercov'] = lisa_fisher.read_h5py_fishercov(hf['/fishercov'])

    # Number of samples
    n = len(post['lnL'])

    # Complete with injected values for fixed parameters
    # NOTE: for multiemcee, post is already complete for fixed params
    # We re-do it here, not optimal
    # NOTE: fixed_params is list_params minus infer_params
    # but list_params is optional in the json; if absent, use source_params
    default_list_params = source_params.copy()
    default_list_params.pop('Lframe', False)
    list_params = prior_params.get('list_params', default_list_params.keys())
    infer_params = prior_params['infer_params']
    fixed_params = [p for p in list_params if p not in infer_params]
    for p in fixed_params:
        if run_params.get('sample_Lframe', False):
            post[p] = np.full(n, source_params_Lframe[p])
        else:
            post[p] = np.full(n, source_params_SSBframe[p])

    # Conversion of the posterior Lfame/SSBframe will happen outside for now
    post['Lframe'] = run_params.get('sample_Lframe', False)

    # Complete mass and spin parameters
    source_params_SSBframe = pytools.complete_mass_params(source_params_SSBframe)
    source_params_SSBframe = pytools.complete_spin_params(source_params_SSBframe)
    source_params_Lframe = pytools.complete_mass_params(source_params_Lframe)
    source_params_Lframe = pytools.complete_spin_params(source_params_Lframe)
    post = pytools.complete_mass_params(post)
    post = pytools.complete_spin_params(post)

    params_post['injparams_SSBframe'] = source_params_SSBframe
    params_post['injparams_Lframe'] = source_params_Lframe
    params_post['waveform_params'] = waveform_params
    params_post['prior_params'] = prior_params
    params_post['run_params'] = run_params
    params_post['post'] = post
    return params_post

# Load posterior file
# Assumed format: param1 param2 ... paramn lnL, parameters read from .json
# Fixed parameters completed with injected value read from .json
def load_params_posterior_lisa_sobh(params_file, posterior_file, format='multiemcee', load_fisher=True, AGN_params=False):

    params_post = {}

    # Load input params
    with open(params_file, 'r') as input_file:
        input_params = json.load(input_file)
    source_params = input_params['source_params']
    waveform_params = input_params['waveform_params']
    run_params = input_params['run_params']
    source_params_phys = source_params.copy()
    source_params_are_Lframe = source_params_phys.pop('Lframe', False)
    prior_params = input_params['prior_params']
    t0 = waveform_params.get('t0', 0.)

    # If source params are given in the Lframe, convert to SSB
    if source_params_are_Lframe:
        raise ValueError('L-frame not relevant/supported for SBHBs.')
    source_params_SSBframe = source_params.copy()

    # Load posterior from file
    if format=='multinest':
        post_raw = np.genfromtxt(posterior_file)
        post = {}
        for i,p in enumerate(prior_params['infer_params']):
            post[p] = post_raw[:,i]
        post['lnL'] = post_raw[:,-1]
    elif format=='multiemcee':
        with h5py.File(posterior_file, 'r') as hf:
            post = {}
            for p in prior_params['infer_params']:
                post[p] = hf[p][:]
            post['lnL'] = hf['lnlike'][:]
            post['lnpost'] = hf['lnpost'][:]
    elif format=='multifisher':
        with h5py.File(posterior_file, 'r') as hf:
            post = {}
            for p in prior_params['infer_params']:
                post[p] = hf[p][:]
            post['lnL'] = hf['lnlike'][:]
            post['lnpost'] = hf['lnpost'][:]
            post['lnweight'] = hf['lnweight'][:]
    else:
        raise ValueError('Format not recognized')

    # For multiemcee, we also load fisher
    if not format=='multinest' and load_fisher:
        with h5py.File(posterior_file, 'r') as hf:
            params_post['fishercov'] = lisa_fisher.read_h5py_fishercov(hf['/fishercov'])

    # Number of samples
    n = len(post['lnL'])

    # Complete with injected values for fixed parameters
    # NOTE: for multiemcee, post is already complete for fixed params
    # We re-do it here, not optimal
    infer_params = prior_params['infer_params']
    fixed_params = [p for p in source_params_phys.keys() if p not in infer_params]
    for p in fixed_params:
        if run_params.get('sample_Lframe', False):
            post[p] = np.full(n, source_params_Lframe[p])
        else:
            post[p] = np.full(n, source_params_SSBframe[p])

    # Conversion of the posterior Lfame/SSBframe will happen outside for now
    post['Lframe'] = run_params.get('sample_Lframe', False)
    if post['Lframe']:
        raise ValueError('L-frame not relevant/supported for SBHBs.')

    # Complete mass and spin parameters
    source_params_SSBframe = pytools.complete_mass_params(source_params_SSBframe)
    source_params_SSBframe = pytools.complete_spin_params(source_params_SSBframe)
    post = pytools.complete_mass_params(post)
    post = pytools.complete_spin_params(post)
    # Only relevant for SOBH_AGN case
    if AGN_params:
        source_params_SSBframe = pytools.complete_AGN_params(source_params_SSBframe)
        post = pytools.complete_AGN_params(post)

    params_post['injparams_SSBframe'] = source_params_SSBframe
    params_post['waveform_params'] = waveform_params
    params_post['prior_params'] = prior_params
    params_post['run_params'] = run_params
    params_post['post'] = post
    return params_post

# Function taken from J.Baker's corner-fisher.py
def read_covariance(file):
    pars=[]
    done=False
    trycount=0
    with open(file,'r') as f:
        line="#"
        while("#" in line): line=f.readline() #Skip comment
        for val in line.split():
            pars.append(float(val))
        Npar=len(pars)
        while(not "#Covariance" in line):line=f.readline() #Skip until the good stuff
        covar=np.zeros((Npar,Npar))
        i=0
        for par in pars:
            line=f.readline()
            covar[i]=np.array(line.split())
            i+=1
    return covar
def read_fisher(file):
    pars=[]
    done=False
    trycount=0
    with open(file,'r') as f:
        line="#"
        while("#" in line): line=f.readline() #Skip comment
        for val in line.split():
            pars.append(float(val))
        Npar=len(pars)
        while(not "#Fisher" in line):line=f.readline() #Skip until the good stuff
        covar=np.zeros((Npar,Npar))
        i=0
        for par in pars:
            line=f.readline()
            covar[i]=np.array(line.split())
            i+=1
    return covar


# Compute posterior values for individual posterior samples by multiplying with the prior
# Assumes a np array for the posterior samples, the last column being the likelihood values
# Assumes flat prior in time, phase, polarization, sphere prior in inclination and sky position
# Normalization is arbitrary
# Parameter order assumed : m1, m2, tRef, dist, phiRef, inc, lambda, beta, pol, loglikelihood
def compute_posterior(posterior, flatdistprior=False, logflatmassprior=False):
    # sort posterior samples, highest loglikelihood first
    posterior = (posterior[posterior[:,9].argsort()])[::-1]
    if flatdistprior and logflatmassprior:
        def prior(x):
            return 1./x[0] * 1./x[1] * np.sin(x[5]) * np.cos(x[7])
    elif flatdistprior and not logflatmassprior:
        def prior(x):
            return np.sin(x[5]) * np.cos(x[7])
    elif not flatdistprior and logflatmassprior:
        def prior(x):
            return 1./x[0] * 1./x[1] * x[3]**2 * np.sin(x[5]) * np.cos(x[7])
    elif not flatdistprior and not logflatmassprior:
        def prior(x):
            return  x[3]**2 * np.sin(x[5]) * np.cos(x[7])
    priorvalues = np.array(list(map(prior, posterior)))
    # normalize (arbitrarily) to the prior value of highest likelihood (injection)
    priorvalues = priorvalues / priorvalues[0]
    posteriorvalues = np.log(priorvalues) + posterior[:,9]
    return np.concatenate((posterior, np.array([posteriorvalues]).T), axis=1)



################################################################################
## Automated parameter labels and scales
################################################################################

# Parameter labels
unit_M_dict = {
    '1e8':r'$(10^{8}\mathrm{M}_{\odot})$',
    '1e7':r'$(10^{7}\mathrm{M}_{\odot})$',
    '1e6':r'$(10^{6}\mathrm{M}_{\odot})$',
    '1e5':r'$(10^{5}\mathrm{M}_{\odot})$',
    '1e4':r'$(10^{4}\mathrm{M}_{\odot})$',
    '1e3':r'$(10^{3}\mathrm{M}_{\odot})$',
    '1e2':r'$(\mathrm{M}_{\odot})$',
    '1e1':r'$(\mathrm{M}_{\odot})$',
    '1':r'$(\mathrm{M}_{\odot})$'}
unit_D_dict = {
    'Gpc':r'$(\mathrm{Gpc})$',
    'Mpc':r'$(\mathrm{Mpc})$'}
unit_t_dict = {
    's':r'$(\mathrm{s})$',
    'ms':r'$(\mathrm{ms})$'}
unit_f_dict = {
    'Hz':r'$(\mathrm{Hz})$',
    'mHz':r'$(\mathrm{mHz})$'}
scale_dict_M = {'1e8':1e8, '1e7':1e7, '1e6':1e6, '1e5':1e5, '1e4':1e4, '1e3':1e3, '1e2':1e2, '1e1':10., '1':1.}
scale_dict_D = {'Gpc':1e3, 'Mpc':1.}
scale_dict_t = {'s':1., 'ms':1e-3}
scale_dict_f = {'Hz':1., 'mHz':1e-3}

# Display of parameter names
def param_label_dict(detector, scales, Lframe=False):
    unit_str_M = unit_M_dict[scales[0]]
    unit_str_D = unit_D_dict[scales[1]]
    unit_str_t = unit_t_dict[scales[2]]
    unit_str_f = unit_f_dict[scales[3]]
    res = {}
    if detector=='LISA':
        res = {
            'm1'     : r'$m_1 \;$' + unit_str_M,
            'm2'     : r'$m_2 \;$' + unit_str_M,
            'chi1'     : r'$\chi_1$',
            'chi2'     : r'$\chi_2$',
            'chis'     : r'$\chi_s$',
            'chia'     : r'$\chi_a$',
            'chip'     : r'$\chi_+$',
            'chim'     : r'$\chi_-$',
            'chiPN'    : r'$\chi_{\rm PN}$',
            'fstart' : r'$f_{\rm start} \;$' + unit_str_f,
            'dist'   : r'$D \;$' + unit_str_D,
            'phi'    : r'$\varphi \; (\mathrm{rad})$',
            'inc'    : r'$\iota \; (\mathrm{rad})$',
            'M'      : r'$M \;$' + unit_str_M,
            'q'      : r'$q$',
            'Mchirp' : r'$\mathcal{M}_c \;$' + unit_str_M,
            'eta'    : r'$\eta$',
            'Omega_AGN': r'$\Omega_{\rm AGN} \; (\mathrm{rad}.\mathrm{s}^{-1})$',
            'Rcostheta_AGN': r'$R/c \cos\theta_{\rm AGN} \; (\mathrm{s})$',
            'Rsintheta_AGN': r'$R/c \sin\theta_{\rm AGN} \; (\mathrm{s})$',
            'phi_AGN': r'$\phi_{\rm AGN} \; (\mathrm{rad})$',
            'R_AGN': r'$R_{\rm AGN} \; (M)$',
            'M_AGN': r'$M_{\rm AGN} \; (M_{\odot})$',
            'theta_AGN': r'$\theta_{\rm AGN} \; (\mathrm{rad})$',
            'Aplus': r'$A_+$',
            'Aminus': r'$A_-$',
            'Phiplus': r'$\Phi_+ \; (\mathrm{rad})$',
            'Phiminus': r'$\Phi_- \; (\mathrm{rad})$',
            'Phis': r'$\Phi_s \; (\mathrm{rad})$',
            'Phia': r'$\Phi_a \; (\mathrm{rad})$',
        }
        if not Lframe:
            res.update({
                'Deltat' : r'$\Delta t \;$' + unit_str_t,
                'lambda' : r'$\lambda \; (\mathrm{rad})$',
                'beta'   : r'$\beta \; (\mathrm{rad})$',
                'psi'    : r'$\psi \; (\mathrm{rad})$',
            })
        else:
            res.update({
                'Deltat' : r'$\Delta t_{L} \;$' + unit_str_t,
                'lambda' : r'$\lambda_{L} \; (\mathrm{rad})$',
                'beta'   : r'$\beta_{L} \; (\mathrm{rad})$',
                'psi'    : r'$\psi_{L} \; (\mathrm{rad})$',
            })
    elif detector=='LLV':
        res = {
            'm1'     : r'$m_1 \;$' + unit_str_M,
            'm2'     : r'$m_2 \;$' + unit_str_M,
            'chi1'     : r'$\chi_1$',
            'chi2'     : r'$\chi_2$',
            'chis'     : r'$\chi_s$',
            'chia'     : r'$\chi_a$',
            'chip'     : r'$\chi_+$',
            'chim'     : r'$\chi_-$',
            'chiPN'    : r'$\chi_{\rm PN}$',
            'Deltat' : r'$\Delta t \;$' + unit_str_t,
            'fstart' : r'$f_{\rm start} \;$' + unit_str_f,
            'dist'   : r'$D \;$' + unit_str_D,
            'phi'    : r'$\varphi \; (\mathrm{rad})$',
            'inc'    : r'$\iota \; (\mathrm{rad})$',
            'ra'     : r'$\mathrm{ra} \; (\mathrm{rad})$',
            'dec'    : r'$\mathrm{dec} \; (\mathrm{rad})$',
            'psi'    : r'$\psi \; (\mathrm{rad})$',
            'M'      : r'$M \;$' + unit_str_M,
            'q'      : r'$q$',
            'Mchirp' : r'$\mathcal{M}_c \;$' + unit_str_M,
            'eta'    : r'$\eta$'
        }
    return res

def automatic_scales(injparams, posterior):
    M = injparams['M']
    scale_M_int = min(8, max(int(np.floor(np.log(M)/np.log(10.))), 0))
    if scale_M_int<=2:
        scale_M = '1'
    else:
        scale_M = '1e' + str(scale_M_int)
    DL = injparams['dist']
    if DL>=1e3:
        scale_D = 'Gpc'
    else:
        scale_D = 'Mpc'
    if 'Deltat' in posterior.keys():
        Deltat_mean = np.abs(np.mean(posterior['Deltat']))
        if Deltat_mean>=0.1:
            scale_t = 's'
        else:
            scale_t = 'ms'
    else:
        scale_t = 's' # Default, not used
    if 'fstart' in posterior.keys():
        scale_fstart = 'mHz'
    else:
        scale_fstart = 'Hz' # Default, not used
    return [scale_M, scale_D, scale_t, scale_fstart]
def scale_posterior(posterior, scales):
    sM, sD, st, sf = scales
    post = copy.deepcopy(posterior)
    # Masses
    post['m1'][:] /= sM
    post['m2'][:] /= sM
    post['M'][:] /= sM
    post['Mchirp'][:] /= sM
    # Distance
    post['dist'][:] /= sD
    # Time
    if 'Deltat' in post.keys():
        post['Deltat'][:] /= st
    # Starting frequency for SBHBs
    if 'fstart' in post.keys():
        post['fstart'][:] /= sf
    return post
def scale_injparams(injparams, scales):
    sM, sD, st, sf = scales
    injpar = injparams.copy()
    # Masses
    injpar['m1'] /= sM
    injpar['m2'] /= sM
    injpar['M'] /= sM
    injpar['Mchirp'] /= sM
    # Distance
    injpar['dist'] /= sD
    # Time
    if 'Deltat' in injpar.keys():
        injpar['Deltat'] /= st
    # Starting frequency for SBHBs
    if 'fstart' in injpar.keys():
        injpar['fstart'] /= sf
    return injpar
# Assumed format for covariance: m1 m2 Deltat D phi inc lambda beta pol
def scale_fishcov(fishcov, scales):
    list_params = fishcov['list_params']
    fishcov_scaled = {}
    fishcov_scaled['params'] = copy.deepcopy(fishcov['params'])
    fishcov_scaled['list_params'] = copy.deepcopy(fishcov['list_params'])
    sM, sD, st, sf = scales
    scale_dict = {'m1':sM, 'm2':sM, 'Mchirp':sM, 'M':sM, 'q':1., 'eta':1., 'chi1':1., 'chi2':1., 'chip':1., 'chim':1., 'chis':1., 'chia':1., 'chiPN':1., 'dist':sD, 'Deltat':st, 'fstart':sf, 'phi':1., 'inc':1., 'lambda':1., 'beta':1., 'psi':1.}
    scaling = np.diag(np.array([1./scale_dict[p] for p in list_params]))
    invscaling = np.diag(np.array([scale_dict[p] for p in list_params]))
    fishcov_scaled['fisher'] = np.dot(scaling, np.dot(fishcov['fisher'], scaling))
    fishcov_scaled['cov'] = np.dot(scaling, np.dot(fishcov['cov'], scaling))
    fishcov_scaled['scales'] = scale_dict
    return fishcov_scaled

################################################################################
## Corner plot function, based on corner_covar, modification of corner.py
################################################################################

# Default levels for contours in 2d histogram - TODO: check the correspondence with 1,2,3sigma
contour_levels_2D_default = 1.0 - np.exp(-0.5 * np.linspace(1.0, 3.0, num=3) ** 2)

# Old default truth color: truth_color="#4682b4"
def corner_plot(injparams, posterior, add_posteriors=None, output=False, output_dir=None, output_file=None, histograms=True, fisher=False, fishercov=None, detector='LISA', params=['m1', 'm2', 'Deltat', 'dist', 'phi', 'inc', 'lambda', 'beta', 'psi'], params_range=None, Lframe=False, lnweight_threshold=-100., labels=None, scales=None, add_truths=None, color="k", add_colors=None, cov_color=None, show_truths=True, truth_color='k', truth_linestyle='-', add_truth_colors=None, add_truth_linestyles=None, bins=50, show_histograms=True, quantiles=[0.159, 0.5, 0.841], show_quantiles=True, levels=contour_levels_2D_default, plot_density=True, plot_contours=True, plot_datapoints=False, smooth=None, smooth1d=None, show_xtick_labels=True, show_ytick_labels=True, label_kwargs={"fontsize": 16}, show_titles=True, titles=None, plot_density_add_post=False, transparent=True, resize_fig=1.):

    # If required, transform to parameters in the L-frame or in the SSB-frame
    if Lframe:
        if not injparams.get('Lframe', False):
            injparams_plt = lisatools.convert_SSBframe_to_Lframe(injparams, t0=injparams['t0'], frozenLISA=False, LISAconst=pyresponse.LISAconstProposal)
        else:
            injparams_plt = injparams
        if not posterior.get('Lframe', False):
            posterior_plt = lisatools.convert_SSBframe_to_Lframe(posterior, t0=injparams['t0'], frozenLISA=False, LISAconst=pyresponse.LISAconstProposal)
        else:
            posterior_plt = posterior
        if add_posteriors is not None:
            add_posteriors_plt = []
            for add_post in add_posteriors:
                if not add_post.get('Lframe', False):
                    add_posteriors_plt += [lisatools.convert_SSBframe_to_Lframe(add_post, t0=injparams['t0'], frozenLISA=False, LISAconst=pyresponse.LISAconstProposal)]
                else:
                    add_posteriors_plt += [add_post]
        else:
            add_posteriors_plt = None
        if add_truths is not None:
            add_truths_plt = []
            for add_truth in add_truths:
                if not add_truth.get('Lframe', False):
                    add_truths_plt += [lisatools.convert_SSBframe_to_Lframe(add_truth, t0=injparams['t0'], frozenLISA=False, LISAconst=pyresponse.LISAconstProposal)]
                else:
                    add_truths_plt += [add_truth]
        else:
            add_truths_plt = None
    else:
        if injparams.get('Lframe', False):
            injparams_plt = lisatools.convert_Lframe_to_SSBframe(injparams, t0=injparams['t0'], frozenLISA=False, LISAconst=pyresponse.LISAconstProposal)
        else:
            injparams_plt = injparams
        if posterior.get('Lframe', False):
            posterior_plt = lisatools.convert_Lframe_to_SSBframe(posterior, t0=injparams['t0'], frozenLISA=False, LISAconst=pyresponse.LISAconstProposal)
        else:
            posterior_plt = posterior
        if add_posteriors is not None:
            add_posteriors_plt = []
            for add_post in add_posteriors:
                if add_post.get('Lframe', False):
                    add_posteriors_plt += [lisatools.convert_Lframe_to_SSBframe(add_post, t0=injparams['t0'], frozenLISA=False, LISAconst=pyresponse.LISAconstProposal)]
                else:
                    add_posteriors_plt += [add_post]
        else:
            add_posteriors_plt = None
        if add_truths is not None:
            add_truths_plt = []
            for add_truth in add_truths:
                if add_truth.get('Lframe', False):
                    add_truths_plt += [lisatools.convert_Lframe_to_SSBframe(add_truth, t0=injparams['t0'], frozenLISA=False, LISAconst=pyresponse.LISAconstProposal)]
                else:
                    add_truths_plt += [add_truth]
        else:
            add_truths_plt = None

    # If not provided, determine automatically scales [M, D, t]
    if scales is None:
        scales = automatic_scales(injparams_plt, posterior_plt)
    scale_M, scale_D, scale_t, scale_f = scales
    scalefactors = scale_dict_M[scale_M], scale_dict_D[scale_D], scale_dict_t[scale_t], scale_dict_f[scale_f]
    injparams_plt = scale_injparams(injparams_plt, scalefactors)
    posterior_plt = scale_posterior(posterior_plt, scalefactors)
    if add_posteriors_plt is not None:
        add_posteriors_plt = list(map(lambda x: scale_posterior(x, scalefactors), add_posteriors))

    # Convert parameters and posteriors from dicts to numpy arrays
    injparams_plt = np.array([injparams_plt[p] for p in params])
    posterior_plt = np.array([posterior_plt[p] for p in params]).T
    if add_posteriors_plt is not None:
        add_posteriors_plt = list(map(lambda x: np.array([x[p] for p in params]).T, add_posteriors_plt))
    if add_truths is not None:
        add_truths = list(map(lambda x: np.array([x[p] for p in params]), add_truths))
    if show_truths:
        truths = injparams_plt
    else:
        truths = None

    # Process weights of individual samples if given
    weights = None
    if 'lnweight' in posterior.keys():
        mask = posterior['lnweight'] > lnweight_threshold
        posterior_plt = posterior_plt[mask]
        weights = np.exp(posterior['lnweight'][mask])
    if add_posteriors is not None:
        add_weights = [None] * len(add_posteriors)
        for i,add_post in enumerate(add_posteriors):
            if 'lnweight' in add_post.keys():
                mask = add_post['lnweight'] > lnweight_threshold
                add_posteriors_plt[i] = add_posteriors_plt[i][mask]
                add_weights[i] = [np.exp(add_post['lnweight'][mask])]

    # Covariance from the Fisher matrix
    if fisher:
        if fishercov is None:
            raise ValueError('Covariance matrix not specified.')
        if Lframe and not fishercov['Lframe']:
            raise ValueError('Covariance matrix not in the L-frame.')
        fishercov = scale_fishcov(fishercov, scalefactors)
        cov_list_params = fishercov['list_params']
        cov_cols = [cov_list_params.index(p) for p in params]
        cov_cols_cartesian = np.ix_(cov_cols, cov_cols)
        cov = fishercov['cov'][cov_cols_cartesian]
    else:
        cov = None

    # Labels
    if labels is None:
        label_dict = param_label_dict(detector, scales, Lframe)
        labels = list(map(lambda x: label_dict[x], params))

    # Main call to corner function
    fig, axes = corner_covar.corner(posterior_plt, cov=cov, bins=bins, params_range=params_range, levels=levels, weights=weights,
                                 labels=labels, label_kwargs=label_kwargs, color=color,
                                 truths=truths, add_truths=add_truths, truth_color=truth_color, truth_linestyle=truth_linestyle, add_truth_colors=add_truth_colors, add_truth_linestyles=add_truth_linestyles, plot_datapoints=plot_datapoints,
                                 smooth=smooth, smooth1d=smooth1d, show_histograms=show_histograms,
                                 quantiles=quantiles, show_quantiles=show_quantiles, plot_contours=plot_contours,
                                 show_xtick_labels=show_xtick_labels, show_ytick_labels=show_ytick_labels,
                                 show_titles=show_titles, titles=titles, plot_density=plot_density, resize_fig=resize_fig)

    # Overlay other posterior
    if add_posteriors_plt is not None:
        if add_colors is None:
            add_colors = [plotpalette[i % len(plotpalette)] for i in range(len(add_posteriors))]
        for i, add_post in enumerate(add_posteriors_plt):
            corner_covar.corner(add_post, figaxes=(fig, axes), cov=None, bins=bins, params_range=params_range, levels=levels, weights=add_weights[i],
                                     labels=labels, label_kwargs=label_kwargs, color=add_colors[i],
                                     truths=None, plot_datapoints=plot_datapoints,
                                     smooth=smooth, smooth1d=smooth1d, show_histograms=show_histograms,
                                     quantiles=quantiles, show_quantiles=show_quantiles, plot_contours=plot_contours,
                                     show_titles=False, titles=titles,
                                     show_xtick_labels=show_xtick_labels, show_ytick_labels=show_ytick_labels,
                                     plot_density=plot_density_add_post, no_fill_contours=True, resize_fig=resize_fig)

    # Output
    if output:
        if not output_dir:
            raise ValueError('output_dir not defined.')
        if not output_file:
            raise ValueError('output_file not defined.')
        fig.savefig(output_dir + output_file, bbox_inches='tight', transparent=transparent)
    else:
        return fig



#
# Copyright (C) 2019 Sylvain Marsat.
#
#


"""
    Python tools for LISA: L-frame, Jacobian, degenerate sky positions.
"""

from __future__ import absolute_import, division, print_function
import sys
if sys.version_info[0] == 2:
    from future_builtins import map, filter


import copy
import numpy as np

import scipy
import scipy.interpolate as ip
from scipy.interpolate import InterpolatedUnivariateSpline as spline

import lisabeta.tools.pytools as pytools
import lisabeta.pyconstants as pyconstants
import lisabeta.lisa.pyresponse as pyresponse

# Constants
R = pyconstants.AU_SI
c = pyconstants.C_SI
Omega = pyconstants.EarthOrbitOmega_SI
msols = pyconstants.MTSUN_SI

################################################################################
# Functions for the L-frame conversion for LISA
################################################################################

# Conversions between SSB-frame and L-frame parameters (for initial position alpha=0)
def functLfromtSSB(tSSB, lambd, beta):
    return tSSB - R/c*np.cos(beta)*np.cos(Omega*tSSB - lambd)
def functSSBfromtL(tL, lambd, beta):
    return tL + R/c*np.cos(beta)*np.cos(Omega*tL - lambd) - 1./2*Omega*(R/c*np.cos(beta))**2*np.sin(2.*(Omega*tL - lambd))
# Option SetphiRefSSBAtfRef indicates the meaning of phase in the input SSB frame
# L-frame phase is always a pure observer phase shift, equivalent to having SetphiRefSSBAtfRef=False AND flipping the sign
# So phiL = -phiSSB(SetphiRefSSBAtfRef=False)
# We set phiL' = -phiSSB(SetphiRefSSBAtfRef=True) + pi*tSSB*fRef to remove the tRef-dependence
# NOTE: phiL != phiL'
# the two still differ by fRef-dependent terms that require access to the waveform to evaluate
# NOTE: phiL' only supports fRef set at its default, the maximum 22 frequency covered by the ROM
# NOTE: fliphase=False is to be used when the SSB phase is already 'correct' (phiRef=phi0 entering the Ylm in code updated from pyFDresponse), in which case we do not touch the phase -- using both SetphiRefSSBAtfRef=True and flipphase=False will not really be use, but we allow it
def funcphiL(m1, m2, tSSB, phiSSB, SetphiRefSSBAtfRef=False, flipphase=True): # note: mod to [0,2pi]
    if not SetphiRefSSBAtfRef:
        if flipphase:
            return -phiSSB
        else:
            return phiSSB
    else:
        MfROMmax22 = 0.14
        fRef = MfROMmax22/((m1 + m2)*msols)
        if flipphase:
            return pytools.mod2pi(-phiSSB + np.pi*tSSB*fRef)
        else:
            return pytools.mod2pi(phiSSB + np.pi*tSSB*fRef)
# Inverse transformation of the phase
# NOTE: we take tSSB as an argument, not tL - because computing tSSB requires the sky position as well
def funcphiSSB(m1, m2, tSSB, phiL, SetphiRefSSBAtfRef=False, flipphase=True): # note: mod to [0,2pi]
    if not SetphiRefSSBAtfRef:
        if flipphase:
            return -phiL
        else:
            return phiL
    else:
        MfROMmax22 = 0.14
        fRef = MfROMmax22/((m1 + m2)*msols)
        if flipphase:
            return pytools.mod2pi(-phiL + np.pi*tSSB*fRef)
        else:
            return pytools.mod2pi(phiL - np.pi*tSSB*fRef)
# NOTE: simple relation between L-frame definitions
# lambdaL_old = lambdaL_paper - pi/2
def funclambdaL(lambd, beta, defLframe='paper'):
    if defLframe=='paper':
        return np.arctan2(np.cos(beta)*np.sin(lambd), np.cos(beta)*np.cos(lambd)*np.cos(np.pi/3) + np.sin(beta)*np.sin(np.pi/3))
    elif defLframe=='old':
        return -np.arctan2(np.cos(beta)*np.cos(lambd)*np.cos(np.pi/3) + np.sin(beta)*np.sin(np.pi/3), np.cos(beta)*np.sin(lambd))
    else:
        raise ValueError('Value %s for defLframe is not recognized.' % defLframe)
def funcbetaL(lambd, beta):
    return -np.arcsin(np.cos(beta)*np.cos(lambd)*np.sin(np.pi/3) - np.sin(beta)*np.cos(np.pi/3))
# NOTE: old equivalent writing
# modpi(np.arctan2(np.cos(np.pi/3)*np.cos(beta)*np.sin(psi) - np.sin(np.pi/3)*(np.sin(lambd)*np.cos(psi) - np.cos(lambd)*np.sin(beta)*np.sin(psi)), np.cos(np.pi/3)*np.cos(beta)*np.cos(psi) + np.sin(np.pi/3)*(np.sin(lambd)*np.sin(psi) + np.cos(lambd)*np.sin(beta)*np.cos(psi))))
def funcpsiL(lambd, beta, psi): # note: mod to [0,pi]
    return pytools.modpi(psi + np.arctan2(-np.sin(np.pi/3)*np.sin(lambd), np.cos(np.pi/3)*np.cos(beta) + np.sin(np.pi/3)*np.cos(lambd)*np.sin(beta)))

# Copy of C functions for translation between frames
# We modify constellation variant, keeping only the initial constellation phase : Omega is fixed, constellation_ini_phase replaces variant->ConstPhi0
# NOTE: C function for time duplicate python functions functLfromtSSB and functSSBfromtL
# Compute Solar System Barycenter time tSSB from retarded time at the center of the LISA constellation tL
# NOTE: depends on the sky position given in SSB parameters
def tSSBfromLframe(tL, lambdaSSB, betaSSB, constellation_ini_phase=0., frozenLISA=False, tfrozenLISA=None):
    if frozenLISA:
        if tfrozenLISA is None:
            alpha = constellation_ini_phase
        else:
            alpha = Omega * (tfrozenLISA) + constellation_ini_phase
    else:
        alpha = Omega * (tL) + constellation_ini_phase
    phase = alpha - lambdaSSB
    RoC = R/c
    if frozenLISA:
        correction_implicit = 0.
    else:
        correction_implicit = -1./2*Omega*pow(RoC*np.cos(betaSSB), 2)*np.sin(2.*phase)
    return tL + RoC*np.cos(betaSSB)*np.cos(phase) + correction_implicit;
# Compute retarded time at the center of the LISA constellation tL from Solar System Barycenter time tSSB */
def tLfromSSBframe(tSSB, lambdaSSB, betaSSB, constellation_ini_phase=0., frozenLISA=False, tfrozenLISA=None):
    if frozenLISA:
        if tfrozenLISA is None:
            alpha = constellation_ini_phase
        else:
            alpha = Omega * (tfrozenLISA) + constellation_ini_phase
    else:
        alpha = Omega * (tSSB) + constellation_ini_phase
    phase = alpha - lambdaSSB
    RoC = R/c
    return tSSB - RoC*np.cos(betaSSB)*np.cos(phase)
# Convert L-frame params to SSB-frame params
# NOTE: no transformation of the phase -- approximant-dependence with e.g. EOBNRv2HMROM setting phiRef at fRef, and freedom in definition
def ConvertLframeParamsToSSBframe(
    tL,
    lambdaL,
    betaL,
    psiL,
    constellation_ini_phase=0.,
    frozenLISA=False,
    tfrozenLISA=None):

    alpha = 0.; cosalpha = 0; sinalpha = 0.; coslambdaL = 0; sinlambdaL = 0.; cosbetaL = 0.; sinbetaL = 0.; cospsiL = 0.; sinpsiL = 0.;
    coszeta = np.cos(np.pi/3.)
    sinzeta = np.sin(np.pi/3.)
    coslambdaL = np.cos(lambdaL)
    sinlambdaL = np.sin(lambdaL)
    cosbetaL = np.cos(betaL)
    sinbetaL = np.sin(betaL)
    cospsiL = np.cos(psiL)
    sinpsiL = np.sin(psiL)
    lambdaSSB_approx = 0.
    betaSSB_approx = 0.
    # Initially, approximate alpha using tL instead of tSSB - then iterate
    # NOTE: iteration is not useful when using frozenLISA - will just repeat
    tSSB_approx = tL
    for k in range(3):
        if frozenLISA:
            if tfrozenLISA is None:
                alpha = constellation_ini_phase
            else:
                alpha = Omega * (tfrozenLISA) + constellation_ini_phase
        else:
            alpha = Omega * (tSSB_approx) + constellation_ini_phase
        cosalpha = np.cos(alpha)
        sinalpha = np.sin(alpha)
        lambdaSSB_approx = np.arctan2(cosalpha*cosalpha*cosbetaL*sinlambdaL -sinalpha*sinbetaL*sinzeta + cosbetaL*coszeta*sinalpha*sinalpha*sinlambdaL -cosalpha*cosbetaL*coslambdaL*sinalpha + cosalpha*cosbetaL*coszeta*coslambdaL*sinalpha, cosbetaL*coslambdaL*sinalpha*sinalpha -cosalpha*sinbetaL*sinzeta + cosalpha*cosalpha*cosbetaL*coszeta*coslambdaL -cosalpha*cosbetaL*sinalpha*sinlambdaL + cosalpha*cosbetaL*coszeta*sinalpha*sinlambdaL)
        betaSSB_approx = np.arcsin(coszeta*sinbetaL + cosalpha*cosbetaL*coslambdaL*sinzeta + cosbetaL*sinalpha*sinzeta*sinlambdaL)
        tSSB_approx = tSSBfromLframe(tL, lambdaSSB_approx, betaSSB_approx, constellation_ini_phase=constellation_ini_phase, frozenLISA=frozenLISA, tfrozenLISA=tfrozenLISA)
    tSSB = tSSB_approx
    lambdaSSB = lambdaSSB_approx
    betaSSB = betaSSB_approx
    # Polarization
    psiSSB = pytools.modpi(psiL + np.arctan2(cosalpha*sinzeta*sinlambdaL -coslambdaL*sinalpha*sinzeta, cosbetaL*coszeta -cosalpha*coslambdaL*sinbetaL*sinzeta -sinalpha*sinbetaL*sinzeta*sinlambdaL))
    return [tSSB, lambdaSSB, betaSSB, psiSSB]
# Convert SSB-frame params to L-frame params
# NOTE: no transformation of the phase -- approximant-dependence with e.g. EOBNRv2HMROM setting phiRef at fRef, and freedom in definition
def ConvertSSBframeParamsToLframe(
    tSSB,
    lambdaSSB,
    betaSSB,
    psiSSB,
    constellation_ini_phase=0.,
    frozenLISA=False,
    tfrozenLISA=None):

    alpha = 0.; cosalpha = 0; sinalpha = 0.; coslambda = 0; sinlambda = 0.; cosbeta = 0.; sinbeta = 0.; cospsi = 0.; sinpsi = 0.;
    coszeta = np.cos(np.pi/3.)
    sinzeta = np.sin(np.pi/3.)
    coslambda = np.cos(lambdaSSB)
    sinlambda = np.sin(lambdaSSB)
    cosbeta = np.cos(betaSSB)
    sinbeta = np.sin(betaSSB)
    cospsi = np.cos(psiSSB)
    sinpsi = np.sin(psiSSB)
    if frozenLISA:
        if tfrozenLISA is None:
            alpha = constellation_ini_phase
        else:
            alpha = Omega * (tfrozenLISA) + constellation_ini_phase
    else:
        alpha = Omega * (tSSB) + constellation_ini_phase
    cosalpha = np.cos(alpha)
    sinalpha = np.sin(alpha)
    tL = tLfromSSBframe(tSSB, lambdaSSB, betaSSB, constellation_ini_phase=constellation_ini_phase, frozenLISA=frozenLISA, tfrozenLISA=tfrozenLISA)
    lambdaL = np.arctan2(cosalpha*cosalpha*cosbeta*sinlambda + sinalpha*sinbeta*sinzeta + cosbeta*coszeta*sinalpha*sinalpha*sinlambda -cosalpha*cosbeta*coslambda*sinalpha + cosalpha*cosbeta*coszeta*coslambda*sinalpha, cosalpha*sinbeta*sinzeta + cosbeta*coslambda*sinalpha*sinalpha + cosalpha*cosalpha*cosbeta*coszeta*coslambda -cosalpha*cosbeta*sinalpha*sinlambda + cosalpha*cosbeta*coszeta*sinalpha*sinlambda)
    betaL = np.arcsin(coszeta*sinbeta -cosalpha*cosbeta*coslambda*sinzeta -cosbeta*sinalpha*sinzeta*sinlambda)
    psiL = pytools.modpi(psiSSB + np.arctan2(coslambda*sinalpha*sinzeta -cosalpha*sinzeta*sinlambda, cosbeta*coszeta + cosalpha*coslambda*sinbeta*sinzeta + sinalpha*sinbeta*sinzeta*sinlambda))
    return [tL, lambdaL, betaL, psiL]
# Wrapper functions for frame conversion, used with dictionary of parameters
# NOTE: t0 in yr
def convert_Lframe_to_SSBframe(params_Lframe, t0=0., frozenLISA=False, LISAconst=pyresponse.LISAconstProposal):
    if not (params_Lframe.get('Lframe', False)):
        raise ValueError('Input params already given in the SSB-frame.')
    t0_s = t0 * pyconstants.YRSID_SI
    tL = params_Lframe['Deltat'] + t0_s
    lambdaL = params_Lframe['lambda']
    betaL = params_Lframe['beta']
    psiL = params_Lframe['psi']
    tSSB, lambdaSSB, betaSSB, psiSSB = ConvertLframeParamsToSSBframe(
        tL,
        lambdaL,
        betaL,
        psiL,
        constellation_ini_phase=0.,
        frozenLISA=frozenLISA,
        tfrozenLISA=t0_s)
    params_SSBframe = copy.deepcopy(params_Lframe)
    params_SSBframe['Deltat'] = tSSB - t0_s
    params_SSBframe['lambda'] = lambdaSSB
    params_SSBframe['beta'] = betaSSB
    params_SSBframe['psi'] = psiSSB
    params_SSBframe['Lframe'] = False
    return params_SSBframe
# NOTE: t0 in yr
def convert_SSBframe_to_Lframe(params_SSBframe, t0=0., frozenLISA=False, LISAconst=pyresponse.LISAconstProposal):
    if(params_SSBframe.get('Lframe', False)):
        raise ValueError('Input params already given in the L-frame.')
    t0_s = t0 * pyconstants.YRSID_SI
    tSSB = params_SSBframe['Deltat'] + t0_s
    lambdaSSB = params_SSBframe['lambda']
    betaSSB = params_SSBframe['beta']
    psiSSB = params_SSBframe['psi']
    tL, lambdaL, betaL, psiL = ConvertSSBframeParamsToLframe(
        tSSB,
        lambdaSSB,
        betaSSB,
        psiSSB,
        constellation_ini_phase=0.,
        frozenLISA=frozenLISA,
        tfrozenLISA=t0_s)
    params_Lframe = copy.deepcopy(params_SSBframe)
    params_Lframe['Deltat'] = tL - t0_s
    params_Lframe['lambda'] = lambdaL
    params_Lframe['beta'] = betaL
    params_Lframe['psi'] = psiL
    params_Lframe['Lframe'] = True
    return params_Lframe
# Compute DeltatL_cut for a given time to merger
# When DeltatSSB is given, we have to compute DeltatL first
# time_to_merger > 0 in s
def DeltatL_cut_from_time_to_merger(time_to_merger, params, **waveform_params):
    if time_to_merger < 0:
        raise ValueError('Must have time_to_merger >= 0, and given in s.')
    Lframe = params.get('Lframe', False)
    # DeltatL (+t0*yr) is the arrival time of the merger at LISA
    if Lframe:
        DeltatL = params['Deltat']
    else:
        DeltatL = tLfromSSBframe(params['Deltat'] + waveform_params['t0']*pyconstants.YRSID_SI, params['lambda'], params['beta']) - waveform_params['t0']*pyconstants.YRSID_SI
    return DeltatL - time_to_merger

################################################################################
# Sky modes in the L-frame
################################################################################

# 8-modes sky degeneracy, index=[a,b] for betaL *=a and lambdaL += b*pi/2
def skymode_degen(params0, index):
    params = params0.copy()
    if not params.get('Lframe', False):
        raise ValueError('Input parameters are not in the L-frame.')
    if index[0]==-1:
        params['beta'] = -params['beta']
        params['inc'] = np.pi - params['inc']
        params['psi'] = np.pi - params['psi']
    elif index[0]==1:
        pass
    else:
        raise ValueError('Invalid sky mode index.')
    params['lambda'] = params['lambda'] + index[1]*np.pi/2
    params['psi'] = params['psi'] + index[1]*np.pi/2
    params['lambda'] = pytools.mod2pi(params['lambda'])
    params['psi'] = pytools.modpi(params['psi'])
    return params

################################################################################
# Derivatives and Jacobian in the SSB-frame L-frame parameter map
################################################################################

# Derivatives of L-frame parameters with respect to SSB-frame parameters
# tL
def funcdtLdtSSB(tSSB, lambd, beta):
    return 1 + R/c * np.cos(beta) * Omega*np.sin(Omega*tSSB - lambd)
def funcdtLdlambda(tSSB, lambd, beta):
    return -R/c * np.cos(beta) * np.sin(Omega*tSSB - lambd)
def funcdtLdbeta(tSSB, lambd, beta):
    return R/c * np.sin(beta) * np.cos(Omega*tSSB - lambd)
# phiL
# Old flare setup: phi was tied to other parameters via phi(fref) = phiref
# Option SetphiRefSSBAtfRef indicates the meaning of phase in the input SSB frame
# L-frame phase is always a pure observer phase phase shift, equivalent to having SetphiRefSSBAtfRef=False AND flipping the sign
# So phiL = -phiSSB(SetphiRefSSBAtfRef=False)
# def funcdphiLdm1(m1, m2, t, phi, SetphiRefSSBAtfRef=False):
#     if SetphiRefSSBAtfRef:
#         MfROMmax22 = 0.14
#         dfRefdm1 = -MfROMmax22/((m1 + m2)**2 * msols)
#         return np.pi*t*dfRefdm1
#     else:
#         return 0.
# def funcdphiLdm2(m1, m2, t, phi, SetphiRefSSBAtfRef=False):
#     if SetphiRefSSBAtfRef:
#         MfROMmax22 = 0.14
#         dfRefdm2 = -MfROMmax22/((m1 + m2)**2 * msols)
#         return np.pi*t*dfRefdm2
#     else:
#         return 0.
# def funcdphiLdt(m1, m2, t, phi, SetphiRefSSBAtfRef=False):
#     if SetphiRefSSBAtfRef:
#         MfROMmax22 = 0.14
#         fRef = MfROMmax22/((m1 + m2)*msols)
#         return np.pi*fRef
#     else:
#         return 0.
# def funcdphiLdphi(m1, m2, t, phi, SetphiRefSSBAtfRef=False):
#     return -1
# lambdaL
# NOTE: derivative identical for different L-frame conventions that differ by a constant shift in lambdaL
def funcdlambdaLdlambda(lambd, beta):
    return (np.cos(beta)*(np.cos(beta)*np.cos(np.pi/3) + np.cos(lambd)*np.sin(beta)*np.sin(np.pi/3))) / ((np.cos(beta)*np.cos(np.pi/3)*np.cos(lambd) + np.sin(beta)*np.sin(np.pi/3))**2 +
 np.cos(beta)**2 * np.sin(lambd)**2)
def funcdlambdaLdbeta(lambd, beta):
    return -((np.sin(np.pi/3)*np.sin(lambd))/((np.cos(beta)*np.cos(np.pi/3)*np.cos(lambd) + np.sin(beta)*np.sin(np.pi/3))**2 +
  np.cos(beta)**2*np.sin(lambd)**2))
# betaL
def funcdbetaLdlambda(lambd, beta):
    return np.cos(beta)*np.sin(np.pi/3)*np.sin(lambd) / np.sqrt(1 - (np.cos(np.pi/3)*np.sin(beta) - np.cos(beta)*np.cos(lambd)*np.sin(np.pi/3))**2)
def funcdbetaLdbeta(lambd, beta):
    return (np.cos(beta)*np.cos(np.pi/3) + np.cos(lambd)*np.sin(beta)*np.sin(np.pi/3)) / np.sqrt(1 - (np.cos(np.pi/3)*np.sin(beta) - np.cos(beta)*np.cos(lambd)*np.sin(np.pi/3))**2)
# psiL
def funcdpsiLdlambda(lambd, beta, psi):
    return -((np.sin(np.pi/3)*(np.cos(beta)*np.cos(np.pi/3)*np.cos(lambd) + np.sin(beta)*np.sin(np.pi/3))) / ((np.cos(beta)*np.cos(np.pi/3) + np.cos(lambd)*np.sin(beta)*np.sin(np.pi/3))**2 + np.sin(np.pi/3)**2 * np.sin(lambd)**2))
def funcdpsiLdbeta(lambd, beta, psi):
    return (np.sin(np.pi/3)*(-np.cos(np.pi/3)*np.sin(beta) + np.cos(beta)*np.cos(lambd)*np.sin(np.pi/3))*np.sin(lambd)) / ((np.cos(beta)*np.cos(np.pi/3) + np.cos(lambd)*np.sin(beta)*np.sin(np.pi/3))**2 +
 np.sin(np.pi/3)**2 * np.sin(lambd)**2)
def funcdpsiLdpsi(lambd, beta, psi):
    return 1.
# Jacobian matrix of the SSB-Lframe transformation
# Required to convert Fisher matrices computed with SSB params to L-frame params
# Parameters order : m1 m2 t D phi inc lambda beta psi
# Matrix Jij = \partial xLi / \partial xj
# In old setup for flare:
# Option SetphiRefSSBAtfRef indicates the meaning of phase in the input SSB frame
# L-frame phase is always a pure phase shift, like having SetphiRefSSBAtfRef=False
# In new setup for lisabeta, no need for option SetphiRefSSBAtfRef
# params is a dictionary, list_params gives the ordering of params
def funcJacobianSSBtoLframe(params, list_params, t0=0.):
    it = list_params.index('Deltat')
    ilambda = list_params.index('lambda')
    ibeta = list_params.index('beta')
    ipsi = list_params.index('psi')
    # SSB parameters
    if params.get('Lframe', False):
        raise ValueError('Input parameters are given in the L-frame.')
    t = params['Deltat'] + t0*pyconstants.YRSID_SI
    lambd = params['lambda']
    beta = params['beta']
    psi = params['psi']
    # Jacobian matrix
    n = len(list_params)
    J = np.identity(n, dtype=float)
    # tL
    J[it,it] = funcdtLdtSSB(t, lambd, beta)
    J[it,ilambda] = funcdtLdlambda(t, lambd, beta)
    J[it,ibeta] = funcdtLdbeta(t, lambd, beta)
    # phi - no interdependency in the lisabeta setup
    # lambdaL
    J[ilambda,ilambda] = funcdlambdaLdlambda(lambd, beta)
    J[ilambda,ibeta] = funcdlambdaLdbeta(lambd, beta)
    # betaL
    J[ibeta,ilambda] = funcdbetaLdlambda(lambd, beta)
    J[ibeta,ibeta] = funcdbetaLdbeta(lambd, beta)
    # psiL
    J[ipsi,ilambda] = funcdpsiLdlambda(lambd, beta, psi)
    J[ipsi,ibeta] = funcdpsiLdbeta(lambd, beta, psi)
    J[ipsi,ipsi] = funcdpsiLdpsi(lambd, beta, psi)
    return J

def funcJacobianMqtom1m2(params, list_params):
    iM = list_params.index('M')
    iq = list_params.index('q')
    # Input mass parameters
    m1 = params['m1']
    m2 = params['m2']
    M = m1 + m2
    q = m1 / m2
    # Jacobian matrix
    n = len(list_params)
    J = np.identity(n, dtype=float)
    J[iM,iM] = q/(1+q)
    J[iq,iM] = 1/(1+q)
    J[iM,iq] = M/(1+q)**2
    J[iq,iq] = -M/(1+q)**2
    list_par = copy.deepcopy(list_params)
    list_par[iM] = 'm1'
    list_par[iq] = 'm2'
    return J, list_par

################################################################################
# Converting Fisher matrices and covariance to different sets of parameters
################################################################################

# Fisher matrix and covariance conversions
# Fisher matrix : if J_{a',b} = \partial xL^a' / \partial x^b Jacobian matrix
# F' = tJ^-1 . F . J^-1
# C' = (F^-1)' = J . C . tJ
# Convert SSB-frame covariance to L-frame
def convert_fisher_covariance_Lframe(fishcov, t0=0., frozenLISA=False,
                                        LISAconst=pyresponse.LISAconstProposal):
    if fishcov['Lframe']:
        raise ValueError('Fisher/Covariance already in the L-frame.')
    fishcov_Lframe = copy.deepcopy(fishcov)
    J = funcJacobianSSBtoLframe(fishcov['params'],
                                fishcov['list_params'], t0=t0)
    Jinv = np.linalg.inv(J)
    fishcov_Lframe['params'] = convert_SSBframe_to_Lframe(fishcov['params'],
                              t0=t0, frozenLISA=frozenLISA, LISAconst=LISAconst)
    fishcov_Lframe['fisher'] = np.dot(Jinv.T, np.dot(fishcov['fisher'], Jinv))
    fishcov_Lframe['cov'] = np.dot(J, np.dot(fishcov['cov'], J.T))
    fishcov_Lframe['Lframe'] = True
    return fishcov_Lframe
# Convert (M,q) to (m1,m2)
def convert_fisher_covariance_Mq_to_m1m2(fishcov):
    fishcov_m1m2 = copy.deepcopy(fishcov)
    J, list_par = funcJacobianMqtom1m2(fishcov['params'], fishcov['list_params'])
    Jinv = np.linalg.inv(J)
    fishcov_m1m2['fisher'] = np.dot(Jinv.T, np.dot(fishcov['fisher'], Jinv))
    fishcov_m1m2['cov'] = np.dot(J, np.dot(fishcov['cov'], J.T))
    fishcov_m1m2['list_params'] = list_par
    return fishcov_m1m2

################################################################################
# Sky area error computed from Fisher matrix
################################################################################

def sky_area_cov(fishercov, sq_deg=False, n_sigma=None, prob=0.90):
    """ Compute sky area error from FiM
    Keyword args:
        sq_deg: convert result to square degrees
        n_sigma: ellipse area corresponds to n_sigma*sigma Gaussian contour
        prob: ellipse area corresponds to the enclosed probability prob
        Note that n_sigma and prob are mutually exclusive.
    """
    beta = fishercov['params']['beta']
    ilambda = fishercov['list_params'].index('lambda')
    ibeta = fishercov['list_params'].index('beta')
    cov = fishercov['cov']
    sigmalambdalambda = cov[ilambda,ilambda]
    sigmabetabeta = cov[ibeta,ibeta]
    sigmalambdabeta = cov[ilambda,ibeta]
    # Determine the prefactor for the desired enclosed probability weight
    # Two options, nb of sigmas or probability, mutually exclusive
    prefactor = 1.
    if not n_sigma is None:
        if not prob is None:
            raise ValueError('n_sigma and prob mutually exclusive.')
        prefactor = n_sigma**2
    else:
        if prob is None:
            raise ValueError('Either one of n_sigma or prob must be set.')
        else:
            prefactor = -2*np.log(1. - prob)
    # In sr
    DeltaOmega = prefactor * np.pi * np.cos(beta) *\
                 np.sqrt(sigmalambdalambda*sigmabetabeta - sigmalambdabeta**2)
    # If asked, convert to sq. deg.
    if sq_deg:
        DeltaOmega = DeltaOmega * (180./np.pi)**2
    return DeltaOmega

################################################################################
# Simplified likelihood
################################################################################

# Pattern functions for A,E channels, see Eq. (43) in arXiv:2003.00357
def Faplus(lambd, beta):
    return 1./2 * (1 + np.sin(beta)**2) * np.cos(2*lambd - np.pi/3)
def Facross(lambd, beta):
    return np.sin(beta) * np.sin(2*lambd - np.pi/3)
def Feplus(lambd, beta):
    return 1./2 * (1 + np.sin(beta)**2) * np.cos(2*lambd + np.pi/6)
def Fecross(lambd, beta):
    return np.sin(beta) * np.sin(2*lambd + np.pi/6)

# From simple_likelihood.py, modified to follow notations of the paper
# Format : pars numpy array in the form [d, phiL, inc, lambdaL, betaL, psiL]
# Here d is DL/DLinj, distance scale relative to the injection
# The angles lambdaL, betaL, psiL have a LISA-frame meaning here - and phiL is the quasi-orbital phase shift defined for the L-frame as well (that is, assuming tL is common to all waveforms, not tSSB)
# factor = 4 int((6 pi f L/c)^(2) |h22|^2) - to be given as input, and gives the SNR scale
# sa, se will be precomputed for the injection
def func_sa_22(params):
    [d, phi, inc, lambd, beta, psi] = params
    Faplusval = Faplus(lambd, beta)
    Facrossval = Facross(lambd, beta)
    a22 = 1./4/d * np.sqrt(5/np.pi) * np.cos(inc/2)**4 * np.exp(2.*1j*(phi-psi)) * (Faplusval + 1j*Facrossval)
    a2m2 = 1./4/d * np.sqrt(5/np.pi) * np.sin(inc/2)**4 * np.exp(2.*1j*(phi+psi)) * (Faplusval - 1j*Facrossval)
    return a22 + a2m2
def func_se_22(params):
    [d, phi, inc, lambd, beta, psi] = params
    Feplusval = Feplus(lambd, beta)
    Fecrossval = Fecross(lambd, beta)
    e22 = 1./4/d * np.sqrt(5/np.pi) * np.cos(inc/2)**4 * np.exp(2.*1j*(phi-psi)) * (Feplusval + 1j*Fecrossval)
    e2m2 = 1./4/d * np.sqrt(5/np.pi) * np.sin(inc/2)**4 * np.exp(2.*1j*(phi+psi)) * (Feplusval - 1j*Fecrossval)
    return e22 + e2m2
def simple_likelihood_22(pars, factor, sainj, seinj):
    return -1./2 * factor * (abs(func_sa_22(pars) - sainj)**2 + abs(func_se_22(pars) - seinj)**2)

# Mode pattern functions, Eq. (44) in arXiv:2003.00357
def Fae_lm(params, modes=None):
    if not params.get('Lframe', False):
        raise ValueError('Need parameters in the L-frame.')
    if modes is None:
        modes = [(2,2)]
    inc = params['inc']
    phi = params['phi']
    lambd = params['lambda']
    beta = params['beta']
    psi = params['psi']
    # A bit slow, 4us each (2us / trigo function ?)
    Faplusval = Faplus(lambd, beta)
    Facrossval = Facross(lambd, beta)
    Feplusval = Feplus(lambd, beta)
    Fecrossval = Fecross(lambd, beta)
    sYlm = np.array([pytools.sYlm(lm[0], lm[1], inc, phi) for lm in modes])
    sYlminusmstar = np.array([np.conj(pytools.sYlm(lm[0], -lm[1], inc, phi)) for lm in modes])
    # (-1)^l -- np.power here is surprisingly slow
    minus1l = np.array([1. - 2*(lm[0] % 2) for lm in modes])
    # Fa_lm = 1./2 * (sYlm * np.exp(-1j*2*psi) * (Faplusval + 1j*Facrossval) + minus1l * sYlminusmstar * np.exp(1j*2*psi) * (Faplusval - 1j*Facrossval))
    # Fe_lm = 1./2 * (sYlm * np.exp(-1j*2*psi) * (Feplusval + 1j*Fecrossval) + minus1l * sYlminusmstar * np.exp(1j*2*psi) * (Feplusval - 1j*Fecrossval))
    # Optimize a bit -- should maybe use a C extension
    e2ipsi = np.exp(1j*2*psi)
    em2ipsi = np.conj(e2ipsi)
    # These simple multiplications are not negligible (should be in C)
    fac_a_mpsi = (1./2 * em2ipsi * (Faplusval + 1j*Facrossval))
    fac_a_ppsi = (1./2 * e2ipsi * (Faplusval - 1j*Facrossval))
    fac_e_mpsi = (1./2 * em2ipsi * (Feplusval + 1j*Fecrossval))
    fac_e_ppsi = (1./2 * e2ipsi * (Feplusval - 1j*Fecrossval))
    Fa_lm = fac_a_mpsi * sYlm + fac_a_ppsi * minus1l * sYlminusmstar
    Fe_lm = fac_e_mpsi * sYlm + fac_e_ppsi * minus1l * sYlminusmstar
    return Fa_lm, Fe_lm

# Simple likelihood with higher modes, see Eqs. (67)-(68) in arXiv:2003.00357
# Params must be in the L-frame, and need injection distance
# No check that the intrinsic params and time match injection values
def sae_lm(params, dist_inj, modes=None):
    dist = params['dist']
    d = dist / dist_inj
    Fa_lm, Fe_lm = Fae_lm(params, modes=modes)
    sa_lm = 1./d * Fa_lm
    se_lm = 1./d * Fe_lm
    return sa_lm, se_lm

# Degenerate sky positions - see PE paper
# Because we take the ratio sigma_+/sigma_-, d and phi are silent
def func_degen_sky(inc, lambd, beta, psi):
    phi_dummy = 0.
    d_dummy = 1.
    params = [d_dummy, phi_dummy, inc, lambd, beta, psi]
    sa = func_sa_22(params)
    se = func_se_22(params)
    sigma_plus = sa + 1j*se
    sigma_minus = sa - 1j*se
    r = sigma_plus / sigma_minus
    lambdaL_star = pytools.mod2pi(np.pi/6. - 1./4 * np.angle(r))
    betaL_star = np.pi/2. - 2*np.arctan((np.abs(r))**(1./4))
    # Resolve the pi/2 degeneracy in lambdaL_star by picking the closest to lambdaL
    lambdaL_star = lambdaL_star + ((lambd - lambdaL_star) - pytools.modsym(lambd - lambdaL_star, np.pi/2))
    return (lambdaL_star, betaL_star)
# Degenerate parameters - see PE paper
# Looks for building degenerate point at 0 inclination and 0 phase
# Gives value of all [d, phi, inc, lambd, beta, psi]
# NOTE: phi=0 might not work with previous waveform conventions where it did not
# have its purely extrinsic meaning
def func_degen_params_0inc_0phase(d, phi, inc, lambd, beta, psi):
    params = [d, phi, inc, lambd, beta, psi]
    sa = func_sa_22(params)
    se = func_se_22(params)
    sigma_plus = 1./2 * (sa + 1j*se)
    sigma_minus = 1./2 * (sa - 1j*se)
    r = sigma_plus / sigma_minus
    # Inclination, phase -- by convention
    inc_star = 0. # we choose to look for face-on point
    phi_star = 0. # we choose to look for zero-phase (exact degen with psi)
    # Sky
    lambdaL_star = pytools.mod2pi(np.pi/6. - 1./4 * np.angle(r))
    betaL_star = np.pi/2. - 2*np.arctan((np.abs(r))**(1./4))
    # Resolve the pi/2 degeneracy in lambdaL_star by picking the closest to lambdaL
    lambdaL_star = lambdaL_star + ((lambd - lambdaL_star) - pytools.modsym(lambd - lambdaL_star, np.pi/2))
    # Distance
    thetaL_star = np.pi/2. - betaL_star
    d_star = 1./4*np.sqrt(5./np.pi) * 1./(1. + np.tan(thetaL_star/2.)**2)**2 / np.abs(sigma_minus)
    # Polarization, for generic phi_star although here we have phi_star = 0
    psiL_star = pytools.modpi(-1./2 * np.angle(sigma_minus) + lambdaL_star - np.pi/6 + phi_star) # defined mod pi
    return (d_star, phi_star, inc_star, lambdaL_star, betaL_star, psiL_star)
# Looks for building degenerate point at 0 inclination
# Gives value of all [d, phi, inc, lambd, beta, psi]
# NOTE: phi=0 might not work with previous waveform conventions where it did not
# have its purely extrinsic meaning
# So if phi_star is None (by default), keep the original value
def func_degen_params_0inc(d, phi, inc, lambd, beta, psi, phi_star=None):
    params = [d, phi, inc, lambd, beta, psi]
    sa = func_sa_22(params)
    se = func_se_22(params)
    sigma_plus = 1./2 * (sa + 1j*se)
    sigma_minus = 1./2 * (sa - 1j*se)
    r = sigma_plus / sigma_minus
    # Inclination, phase -- by convention
    inc_star = 0. # we choose to look for face-on point
    if phi_star is None:
        phi_star = phi # by default, just keep the same value of the phase
    # Sky
    lambdaL_star = pytools.mod2pi(np.pi/6. - 1./4 * np.angle(r))
    betaL_star = np.pi/2. - 2*np.arctan((np.abs(r))**(1./4))
    # Resolve the pi/2 degeneracy in lambdaL_star by picking the closest to lambdaL
    lambdaL_star = lambdaL_star + ((lambd - lambdaL_star) - pytools.modsym(lambd - lambdaL_star, np.pi/2))
    # Distance
    thetaL_star = np.pi/2. - betaL_star
    d_star = 1./4*np.sqrt(5./np.pi) * 1./(1. + np.tan(thetaL_star/2.)**2)**2 / np.abs(sigma_minus)
    # Polarization, for generic phi_star although here we have phi_star = 0
    psiL_star = pytools.modpi(-1./2 * np.angle(sigma_minus) + lambdaL_star - np.pi/6 + phi_star) # defined mod pi
    return (d_star, phi_star, inc_star, lambdaL_star, betaL_star, psiL_star)

################################################################################
# Function to compute likelihood for the 8 sky modes
################################################################################

# From a likelihood class, compute the loglikelihood for the 8 sky modes
# Sky indices: (a,b) a=+-1 for a*betaL, b=0-3 for lambdaL + b*pi/2
def func_loglikelihood_skymodes(likelihoodClass):

    params = likelihoodClass.source_params_Lframe
    inc = params['inc']
    lambdaL = params['lambda']
    betaL = params['beta']
    psiL = params['psi']

    skymodes = [(1,0), (-1,0), (1,1), (-1,1), (1,2), (-1,2), (1,3), (-1,3)]

    lnL_skymodes = {}
    for skymode in skymodes:
        params_skymode = params.copy()
        skymode_index0, skymode_index1 = skymode
        params_skymode['inc'] = pytools.modpi(np.pi/2 - (skymode_index0 *(np.pi/2 - inc)))
        params_skymode['lambda'] = pytools.mod2pi(lambdaL + skymode_index1 * np.pi/2)
        params_skymode['beta'] = skymode_index0 * betaL
        params_skymode['psi'] = pytools.modpi(np.pi/2 - (skymode_index0 *(np.pi/2 - psiL)) + skymode_index1 * np.pi/2)
        lnL_skymodes[skymode] = likelihoodClass.lnL(params_skymode, **likelihoodClass.waveform_params)

    return lnL_skymodes
